#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = '''
---
module: kubevirt_rs

short_description: Manage KubeVirt virtual machine replica sets

description:
    - Use Openshift Python SDK to manage the state of KubeVirt virtual machine replica sets.


author: KubeVirt Team (@kubevirt)

options:
    state:
        description:
            - Create or delete virtual machine replica sets.
        default: "present"
        choices:
            - present
            - absent
        type: str
    name:
        description:
            - Name of the virtual machine replica set.
        required: true
        type: str
    namespace:
        description:
            - Namespace where the virtual machine replica set exists.
        required: true
        type: str
    selector:
        description:
            - "Selector is a label query over a set of virtual machine."
        required: true
        type: dict
    replicas:
        description:
            - Number of desired pods. This is a pointer to distinguish between explicit zero and not specified.
            - Replicas defaults to 1 if newly created replica set.
        type: int

extends_documentation_fragment:
- community.kubernetes.k8s_auth_options
- community.general.kubevirt_vm_options
- community.general.kubevirt_common_options


requirements:
  - python >= 2.7
  - openshift >= 0.8.2
'''

EXAMPLES = '''
- name: Create virtual machine replica set 'myvmir'
  community.general.kubevirt_rs:
      state: present
      name: myvmir
      namespace: vms
      wait: true
      replicas: 3
      memory: 64M
      labels:
        myvmi: myvmi
      selector:
        matchLabels:
            myvmi: myvmi
      disks:
         - name: containerdisk
           volume:
             containerDisk:
               image: kubevirt/cirros-container-disk-demo:latest
               path: /custom-disk/cirros.img
           disk:
             bus: virtio

- name: Remove virtual machine replica set 'myvmir'
  community.general.kubevirt_rs:
      state: absent
      name: myvmir
      namespace: vms
      wait: true
'''

RETURN = '''
kubevirt_rs:
  description:
    - The virtual machine virtual machine replica set managed by the user.
    - "This dictionary contains all values returned by the KubeVirt API all options
       are described here U(https://kubevirt.io/api-reference/master/definitions.html#_v1_virtualmachineinstance)"
  returned: success
  type: complex
  contains: {}
'''

import copy
import traceback


from ansible_collections.community.kubernetes.plugins.module_utils.common import AUTH_ARG_SPEC

from ansible_collections.community.general.plugins.module_utils.kubevirt import (
    virtdict,
    KubeVirtRawModule,
    VM_COMMON_ARG_SPEC,
)


KIND = 'VirtualMachineInstanceReplicaSet'
VMIR_ARG_SPEC = {
    'replicas': {'type': 'int'},
    'selector': {'type': 'dict'},
}


class KubeVirtVMIRS(KubeVirtRawModule):

    @property
    def argspec(self):
        """ argspec property builder """
        argument_spec = copy.deepcopy(AUTH_ARG_SPEC)
        argument_spec.update(copy.deepcopy(VM_COMMON_ARG_SPEC))
        argument_spec.update(copy.deepcopy(VMIR_ARG_SPEC))
        return argument_spec

    def wait_for_replicas(self, replicas):
        """ Wait for ready_replicas to equal the requested number of replicas. """
        resource = self.find_supported_resource(KIND)
        return_obj = None

        for event in resource.watch(namespace=self.namespace, timeout=self.params.get('wait_timeout')):
            entity = event['object']
            if entity.metadata.name != self.name:
                continue
            status = entity.get('status', {})
            readyReplicas = status.get('readyReplicas', 0)
            if readyReplicas == replicas:
                return_obj = entity
                break

        if not return_obj:
            self.fail_json(msg="Error fetching the patched object. Try a higher wait_timeout value.")
        if replicas and return_obj.status.readyReplicas is None:
            self.fail_json(msg="Failed to fetch the number of ready replicas. Try a higher wait_timeout value.")
        if replicas and return_obj.status.readyReplicas != replicas:
            self.fail_json(msg="Number of ready replicas is {0}. Failed to reach {1} ready replicas within "
                               "the wait_timeout period.".format(return_obj.status.ready_replicas, replicas))
        return return_obj.to_dict()

    def execute_module(self):
        # Parse parameters specific for this module:
        definition = virtdict()
        selector = self.params.get('selector')
        replicas = self.params.get('replicas')

        if selector:
            definition['spec']['selector'] = selector

        if replicas is not None:
            definition['spec']['replicas'] = replicas

        # defaults for template
        defaults = {'disks': [], 'volumes': [], 'interfaces': [], 'networks': []}

        # Execute the CURD of VM:
        template = definition['spec']['template']
        dummy, definition = self.construct_vm_definition(KIND, definition, template, defaults)
        result_crud = self.execute_crud(KIND, definition)
        changed = result_crud['changed']
        result = result_crud.pop('result')

        # When creating a new VMIRS object without specifying `replicas`, assume it's '1' to make the
        # wait logic work correctly
        if changed and result_crud['method'] == 'create' and replicas is None:
            replicas = 1

        # Wait for the new number of ready replicas after a CRUD update
        # Note1: doesn't work correctly when reducing number of replicas due to how VMIRS works (as of kubevirt 1.5.0)
        # Note2: not the place to wait for the VMIs to get deleted when deleting the VMIRS object; that *might* be
        #        achievable in execute_crud(); keywords: orphanDependents, propagationPolicy, DeleteOptions
        if self.params.get('wait') and replicas is not None and self.params.get('state') == 'present':
            result = self.wait_for_replicas(replicas)

        # Return from the module:
        self.exit_json(**{
            'changed': changed,
            'kubevirt_rs': result,
            'result': result_crud,
        })


def main():
    module = KubeVirtVMIRS()
    try:
        module.execute_module()
    except Exception as e:
        module.fail_json(msg=str(e), exception=traceback.format_exc())


if __name__ == '__main__':
    main()



#!/usr/bin/python
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


DOCUMENTATION = '''
module: aws_region_info
short_description: Gather information about AWS regions.
version_added: 1.0.0
description:
    - Gather information about AWS regions.
    - This module was called C(aws_region_facts) before Ansible 2.9. The usage did not change.
author: 'Henrique Rodrigues (@Sodki)'
options:
  filters:
    description:
      - A dict of filters to apply.
      - Each dict item consists of a filter key and a filter value.
      - See U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeRegions.html) for possible filters.
      - Filter names and values are case sensitive.
      - You can use underscores instead of dashes (-) in the filter keys.
      - Filter keys with underscores will take precedence in case of conflict.
    default: {}
    type: dict
extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2

requirements: [botocore, boto3]
'''

EXAMPLES = '''
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Gather information about all regions
- community.aws.aws_region_info:

# Gather information about a single region
- community.aws.aws_region_info:
    filters:
      region-name: eu-west-1
'''

RETURN = '''
regions:
    returned: on success
    description: >
        Regions that match the provided filters. Each element consists of a dict with all the information related
        to that region.
    type: list
    sample: "[{
        'endpoint': 'ec2.us-west-1.amazonaws.com',
        'region_name': 'us-west-1'
    }]"
'''

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry, ansible_dict_to_boto3_filter_list, camel_dict_to_snake_dict

try:
    from botocore.exceptions import ClientError, BotoCoreError
except ImportError:
    pass  # Handled by AnsibleAWSModule


def main():
    argument_spec = dict(
        filters=dict(default={}, type='dict')
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)
    if module._name == 'aws_region_facts':
        module.deprecate("The 'aws_region_facts' module has been renamed to 'aws_region_info'", date='2021-12-01', collection_name='community.aws')

    connection = module.client('ec2', retry_decorator=AWSRetry.jittered_backoff())

    # Replace filter key underscores with dashes, for compatibility
    sanitized_filters = dict(module.params.get('filters'))
    for k in module.params.get('filters').keys():
        if "_" in k:
            sanitized_filters[k.replace('_', '-')] = sanitized_filters[k]
            del sanitized_filters[k]

    try:
        regions = connection.describe_regions(
            Filters=ansible_dict_to_boto3_filter_list(sanitized_filters)
        )
    except (BotoCoreError, ClientError) as e:
        module.fail_json_aws(e, msg="Unable to describe regions.")

    module.exit_json(regions=[camel_dict_to_snake_dict(r) for r in regions['Regions']])


if __name__ == '__main__':
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the resource
#   module builder playbook.
#
# Do not edit this file manually.
#
# Changes to this file will be over written
#   by the resource module builder.
#
# Changes should be made in the model used to
#   generate this file or in the resource module
#   builder template.
#
#############################################

from __future__ import absolute_import, division, print_function

__metaclass__ = type

"""
The arg spec for the nxos_ospfv2 module
"""


class Ospfv2Args(object):  # pylint: disable=R0903
    """The arg spec for the nxos_ospfv2 module
    """

    def __init__(self, **kwargs):
        pass

    argument_spec = {
        "running_config": {"type": "str"},
        "config": {
            "options": {
                "processes": {
                    "elements": "dict",
                    "options": {
                        "areas": {
                            "required_one_of": [
                                [
                                    "authentication",
                                    "default_cost",
                                    "filter_list",
                                    "nssa",
                                    "ranges",
                                    "stub",
                                ]
                            ],
                            "elements": "dict",
                            "options": {
                                "area_id": {"type": "str", "required": True},
                                "authentication": {
                                    "options": {
                                        "set": {"type": "bool"},
                                        "message_digest": {"type": "bool"},
                                    },
                                    "type": "dict",
                                },
                                "default_cost": {"type": "int"},
                                "filter_list": {
                                    "options": {
                                        "direction": {
                                            "choices": ["in", "out"],
                                            "type": "str",
                                            "required": True,
                                        },
                                        "route_map": {
                                            "type": "str",
                                            "required": True,
                                        },
                                    },
                                    "type": "list",
                                    "elements": "dict",
                                },
                                "nssa": {
                                    "options": {
                                        "default_information_originate": {
                                            "type": "bool"
                                        },
                                        "no_redistribution": {"type": "bool"},
                                        "no_summary": {"type": "bool"},
                                        "set": {"type": "bool"},
                                        "translate": {
                                            "options": {
                                                "type7": {
                                                    "mutually_exclusive": [
                                                        ["always", "never"]
                                                    ],
                                                    "options": {
                                                        "always": {
                                                            "type": "bool"
                                                        },
                                                        "never": {
                                                            "type": "bool"
                                                        },
                                                        "supress_fa": {
                                                            "type": "bool"
                                                        },
                                                    },
                                                    "type": "dict",
                                                }
                                            },
                                            "type": "dict",
                                        },
                                    },
                                    "type": "dict",
                                },
                                "ranges": {
                                    "elements": "dict",
                                    "options": {
                                        "cost": {"type": "int"},
                                        "not_advertise": {"type": "bool"},
                                        "prefix": {
                                            "type": "str",
                                            "required": True,
                                        },
                                    },
                                    "type": "list",
                                },
                                "stub": {
                                    "options": {
                                        "no_summary": {"type": "bool"},
                                        "set": {"type": "bool"},
                                    },
                                    "type": "dict",
                                },
                            },
                            "type": "list",
                        },
                        "auto_cost": {
                            "options": {
                                "reference_bandwidth": {
                                    "type": "int",
                                    "required": True,
                                },
                                "unit": {
                                    "choices": ["Gbps", "Mbps"],
                                    "type": "str",
                                    "required": True,
                                },
                            },
                            "type": "dict",
                        },
                        "bfd": {"type": "bool"},
                        "default_information": {
                            "options": {
                                "originate": {
                                    "options": {
                                        "always": {"type": "bool"},
                                        "route_map": {"type": "str"},
                                        "set": {"type": "bool"},
                                    },
                                    "type": "dict",
                                }
                            },
                            "type": "dict",
                        },
                        "default_metric": {"type": "int"},
                        "distance": {"type": "int"},
                        "flush_routes": {"type": "bool"},
                        "graceful_restart": {
                            "options": {
                                "grace_period": {"type": "int"},
                                "helper_disable": {"type": "bool"},
                                "set": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                        "isolate": {"type": "bool"},
                        "log_adjacency_changes": {
                            "options": {
                                "detail": {"type": "bool"},
                                "log": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                        "max_lsa": {
                            "options": {
                                "ignore_count": {"type": "int"},
                                "ignore_time": {"type": "int"},
                                "max_non_self_generated_lsa": {
                                    "type": "int",
                                    "required": True,
                                },
                                "reset_time": {"type": "int"},
                                "threshold": {"type": "int"},
                                "warning_only": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                        "max_metric": {
                            "options": {
                                "router_lsa": {
                                    "options": {
                                        "external_lsa": {
                                            "options": {
                                                "max_metric_value": {
                                                    "type": "int"
                                                },
                                                "set": {"type": "bool"},
                                            },
                                            "type": "dict",
                                        },
                                        "include_stub": {"type": "bool"},
                                        "on_startup": {
                                            "options": {
                                                "set": {"type": "bool"},
                                                "wait_for_bgp_asn": {
                                                    "type": "int"
                                                },
                                                "wait_period": {"type": "int"},
                                            },
                                            "type": "dict",
                                        },
                                        "set": {"type": "bool"},
                                        "summary_lsa": {
                                            "options": {
                                                "max_metric_value": {
                                                    "type": "int"
                                                },
                                                "set": {"type": "bool"},
                                            },
                                            "type": "dict",
                                        },
                                    },
                                    "type": "dict",
                                }
                            },
                            "type": "dict",
                        },
                        "maximum_paths": {"type": "int"},
                        "mpls": {
                            "options": {
                                "traffic_eng": {
                                    "options": {
                                        "areas": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "area_id": {"type": "str"}
                                            },
                                        },
                                        "multicast_intact": {"type": "bool"},
                                        "router_id": {"type": "str"},
                                    },
                                    "type": "dict",
                                }
                            },
                            "type": "dict",
                        },
                        "name_lookup": {"type": "bool"},
                        "passive_interface": {
                            "options": {"default": {"type": "bool"}},
                            "type": "dict",
                        },
                        "process_id": {"required": True, "type": "str"},
                        "redistribute": {
                            "elements": "dict",
                            "options": {
                                "id": {"type": "str"},
                                "protocol": {
                                    "choices": [
                                        "bgp",
                                        "direct",
                                        "eigrp",
                                        "isis",
                                        "lisp",
                                        "ospf",
                                        "rip",
                                        "static",
                                    ],
                                    "required": True,
                                    "type": "str",
                                },
                                "route_map": {"type": "str", "required": True},
                            },
                            "type": "list",
                        },
                        "rfc1583compatibility": {"type": "bool"},
                        "router_id": {"type": "str"},
                        "shutdown": {"type": "bool"},
                        "summary_address": {
                            "elements": "dict",
                            "mutually_exclusive": [["not_advertise", "tag"]],
                            "options": {
                                "not_advertise": {"type": "bool"},
                                "prefix": {"type": "str", "required": True},
                                "tag": {"type": "int"},
                            },
                            "type": "list",
                        },
                        "table_map": {
                            "options": {
                                "filter": {"type": "bool"},
                                "name": {"type": "str", "required": True},
                            },
                            "type": "dict",
                        },
                        "timers": {
                            "options": {
                                "lsa_arrival": {"type": "int"},
                                "lsa_group_pacing": {"type": "int"},
                                "throttle": {
                                    "options": {
                                        "lsa": {
                                            "options": {
                                                "hold_interval": {
                                                    "type": "int"
                                                },
                                                "max_interval": {
                                                    "type": "int"
                                                },
                                                "start_interval": {
                                                    "type": "int"
                                                },
                                            },
                                            "type": "dict",
                                        },
                                        "spf": {
                                            "options": {
                                                "initial_spf_delay": {
                                                    "type": "int"
                                                },
                                                "max_wait_time": {
                                                    "type": "int"
                                                },
                                                "min_hold_time": {
                                                    "type": "int"
                                                },
                                            },
                                            "type": "dict",
                                        },
                                    },
                                    "type": "dict",
                                },
                            },
                            "type": "dict",
                        },
                        "vrfs": {
                            "elements": "dict",
                            "options": {
                                "areas": {
                                    "required_one_of": [
                                        [
                                            "authentication",
                                            "default_cost",
                                            "filter_list",
                                            "nssa",
                                            "ranges",
                                            "stub",
                                        ]
                                    ],
                                    "elements": "dict",
                                    "options": {
                                        "area_id": {
                                            "type": "str",
                                            "required": True,
                                        },
                                        "authentication": {
                                            "options": {
                                                "set": {"type": "bool"},
                                                "message_digest": {
                                                    "type": "bool"
                                                },
                                            },
                                            "type": "dict",
                                        },
                                        "default_cost": {"type": "int"},
                                        "filter_list": {
                                            "options": {
                                                "direction": {
                                                    "choices": ["in", "out"],
                                                    "type": "str",
                                                    "required": True,
                                                },
                                                "route_map": {
                                                    "type": "str",
                                                    "required": True,
                                                },
                                            },
                                            "type": "list",
                                            "elements": "dict",
                                        },
                                        "nssa": {
                                            "options": {
                                                "default_information_originate": {
                                                    "type": "bool"
                                                },
                                                "no_redistribution": {
                                                    "type": "bool"
                                                },
                                                "no_summary": {"type": "bool"},
                                                "set": {"type": "bool"},
                                                "translate": {
                                                    "options": {
                                                        "type7": {
                                                            "mutually_exclusive": [
                                                                [
                                                                    "always",
                                                                    "never",
                                                                ]
                                                            ],
                                                            "options": {
                                                                "always": {
                                                                    "type": "bool"
                                                                },
                                                                "never": {
                                                                    "type": "bool"
                                                                },
                                                                "supress_fa": {
                                                                    "type": "bool"
                                                                },
                                                            },
                                                            "type": "dict",
                                                        }
                                                    },
                                                    "type": "dict",
                                                },
                                            },
                                            "type": "dict",
                                        },
                                        "ranges": {
                                            "elements": "dict",
                                            "options": {
                                                "cost": {"type": "int"},
                                                "not_advertise": {
                                                    "type": "bool"
                                                },
                                                "prefix": {
                                                    "type": "str",
                                                    "required": True,
                                                },
                                            },
                                            "type": "list",
                                        },
                                        "stub": {
                                            "options": {
                                                "no_summary": {"type": "bool"},
                                                "set": {"type": "bool"},
                                            },
                                            "type": "dict",
                                        },
                                    },
                                    "type": "list",
                                },
                                "auto_cost": {
                                    "options": {
                                        "reference_bandwidth": {
                                            "type": "int",
                                            "required": True,
                                        },
                                        "unit": {
                                            "choices": ["Gbps", "Mbps"],
                                            "type": "str",
                                            "required": True,
                                        },
                                    },
                                    "type": "dict",
                                },
                                "bfd": {"type": "bool"},
                                "capability": {
                                    "type": "dict",
                                    "options": {
                                        "vrf_lite": {
                                            "type": "dict",
                                            "options": {
                                                "set": {"type": "bool"},
                                                "evpn": {"type": "bool"},
                                            },
                                        }
                                    },
                                },
                                "default_information": {
                                    "options": {
                                        "originate": {
                                            "options": {
                                                "always": {"type": "bool"},
                                                "route_map": {"type": "str"},
                                                "set": {"type": "bool"},
                                            },
                                            "type": "dict",
                                        }
                                    },
                                    "type": "dict",
                                },
                                "default_metric": {"type": "int"},
                                "distance": {"type": "int"},
                                "down_bit_ignore": {"type": "bool"},
                                "graceful_restart": {
                                    "options": {
                                        "grace_period": {"type": "int"},
                                        "helper_disable": {"type": "bool"},
                                        "set": {"type": "bool"},
                                    },
                                    "type": "dict",
                                },
                                "log_adjacency_changes": {
                                    "options": {
                                        "detail": {"type": "bool"},
                                        "log": {"type": "bool"},
                                    },
                                    "type": "dict",
                                },
                                "max_lsa": {
                                    "options": {
                                        "ignore_count": {"type": "int"},
                                        "ignore_time": {"type": "int"},
                                        "max_non_self_generated_lsa": {
                                            "type": "int",
                                            "required": True,
                                        },
                                        "reset_time": {"type": "int"},
                                        "threshold": {"type": "int"},
                                        "warning_only": {"type": "bool"},
                                    },
                                    "type": "dict",
                                },
                                "max_metric": {
                                    "options": {
                                        "router_lsa": {
                                            "options": {
                                                "external_lsa": {
                                                    "options": {
                                                        "max_metric_value": {
                                                            "type": "int"
                                                        },
                                                        "set": {
                                                            "type": "bool"
                                                        },
                                                    },
                                                    "type": "dict",
                                                },
                                                "include_stub": {
                                                    "type": "bool"
                                                },
                                                "on_startup": {
                                                    "options": {
                                                        "set": {
                                                            "type": "bool"
                                                        },
                                                        "wait_for_bgp_asn": {
                                                            "type": "int"
                                                        },
                                                        "wait_period": {
                                                            "type": "int"
                                                        },
                                                    },
                                                    "type": "dict",
                                                },
                                                "set": {"type": "bool"},
                                                "summary_lsa": {
                                                    "options": {
                                                        "max_metric_value": {
                                                            "type": "int"
                                                        },
                                                        "set": {
                                                            "type": "bool"
                                                        },
                                                    },
                                                    "type": "dict",
                                                },
                                            },
                                            "type": "dict",
                                        }
                                    },
                                    "type": "dict",
                                },
                                "maximum_paths": {"type": "int"},
                                "name_lookup": {"type": "bool"},
                                "passive_interface": {
                                    "options": {"default": {"type": "bool"}},
                                    "type": "dict",
                                },
                                "redistribute": {
                                    "elements": "dict",
                                    "options": {
                                        "id": {"type": "str"},
                                        "protocol": {
                                            "choices": [
                                                "bgp",
                                                "direct",
                                                "eigrp",
                                                "isis",
                                                "lisp",
                                                "ospf",
                                                "rip",
                                                "static",
                                            ],
                                            "required": True,
                                            "type": "str",
                                        },
                                        "route_map": {
                                            "type": "str",
                                            "required": True,
                                        },
                                    },
                                    "type": "list",
                                },
                                "rfc1583compatibility": {"type": "bool"},
                                "router_id": {"type": "str"},
                                "shutdown": {"type": "bool"},
                                "summary_address": {
                                    "elements": "dict",
                                    "options": {
                                        "not_advertise": {"type": "bool"},
                                        "prefix": {
                                            "type": "str",
                                            "required": True,
                                        },
                                        "tag": {"type": "int"},
                                    },
                                    "type": "list",
                                },
                                "table_map": {
                                    "options": {
                                        "filter": {"type": "bool"},
                                        "name": {
                                            "type": "str",
                                            "required": True,
                                        },
                                    },
                                    "type": "dict",
                                },
                                "timers": {
                                    "options": {
                                        "lsa_arrival": {"type": "int"},
                                        "lsa_group_pacing": {"type": "int"},
                                        "throttle": {
                                            "options": {
                                                "lsa": {
                                                    "options": {
                                                        "hold_interval": {
                                                            "type": "int"
                                                        },
                                                        "max_interval": {
                                                            "type": "int"
                                                        },
                                                        "start_interval": {
                                                            "type": "int"
                                                        },
                                                    },
                                                    "type": "dict",
                                                },
                                                "spf": {
                                                    "options": {
                                                        "initial_spf_delay": {
                                                            "type": "int"
                                                        },
                                                        "max_wait_time": {
                                                            "type": "int"
                                                        },
                                                        "min_hold_time": {
                                                            "type": "int"
                                                        },
                                                    },
                                                    "type": "dict",
                                                },
                                            },
                                            "type": "dict",
                                        },
                                    },
                                    "type": "dict",
                                },
                                "vrf": {"required": True, "type": "str"},
                            },
                            "type": "list",
                        },
                    },
                    "type": "list",
                }
            },
            "type": "dict",
        },
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "gathered",
                "rendered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = '''
---
module: kubevirt_preset

short_description: Manage KubeVirt virtual machine presets

description:
    - Use Openshift Python SDK to manage the state of KubeVirt virtual machine presets.


author: KubeVirt Team (@kubevirt)

options:
    state:
        description:
            - Create or delete virtual machine presets.
        default: "present"
        choices:
            - present
            - absent
        type: str
    name:
        description:
            - Name of the virtual machine preset.
        required: true
        type: str
    namespace:
        description:
            - Namespace where the virtual machine preset exists.
        required: true
        type: str
    selector:
        description:
            - "Selector is a label query over a set of virtual machine preset."
        type: dict

extends_documentation_fragment:
- community.kubernetes.k8s_auth_options
- community.general.kubevirt_vm_options
- community.general.kubevirt_common_options


requirements:
  - python >= 2.7
  - openshift >= 0.8.2
'''

EXAMPLES = '''
- name: Create virtual machine preset 'vmi-preset-small'
  community.general.kubevirt_preset:
      state: present
      name: vmi-preset-small
      namespace: vms
      memory: 64M
      selector:
        matchLabels:
            kubevirt.io/vmPreset: vmi-preset-small

- name: Remove virtual machine preset 'vmi-preset-small'
  community.general.kubevirt_preset:
      state: absent
      name: vmi-preset-small
      namespace: vms
'''

RETURN = '''
kubevirt_preset:
  description:
    - The virtual machine preset managed by the user.
    - "This dictionary contains all values returned by the KubeVirt API all options
       are described here U(https://kubevirt.io/api-reference/master/definitions.html#_v1_virtualmachineinstancepreset)"
  returned: success
  type: complex
  contains: {}
'''

import copy
import traceback


from ansible_collections.community.kubernetes.plugins.module_utils.common import AUTH_ARG_SPEC

from ansible_collections.community.general.plugins.module_utils.kubevirt import (
    virtdict,
    KubeVirtRawModule,
    VM_COMMON_ARG_SPEC
)


KIND = 'VirtualMachineInstancePreset'
VMP_ARG_SPEC = {
    'selector': {'type': 'dict'},
}


class KubeVirtVMPreset(KubeVirtRawModule):

    @property
    def argspec(self):
        """ argspec property builder """
        argument_spec = copy.deepcopy(AUTH_ARG_SPEC)
        argument_spec.update(VM_COMMON_ARG_SPEC)
        argument_spec.update(VMP_ARG_SPEC)
        return argument_spec

    def execute_module(self):
        # Parse parameters specific for this module:
        definition = virtdict()
        selector = self.params.get('selector')

        if selector:
            definition['spec']['selector'] = selector

        # FIXME: Devices must be set, but we don't yet support any
        # attributes there, remove when we do:
        definition['spec']['domain']['devices'] = dict()

        # defaults for template
        defaults = {'disks': [], 'volumes': [], 'interfaces': [], 'networks': []}

        # Execute the CURD of VM:
        dummy, definition = self.construct_vm_definition(KIND, definition, definition, defaults)
        result_crud = self.execute_crud(KIND, definition)
        changed = result_crud['changed']
        result = result_crud.pop('result')

        # Return from the module:
        self.exit_json(**{
            'changed': changed,
            'kubevirt_preset': result,
            'result': result_crud,
        })


def main():
    module = KubeVirtVMPreset()
    try:
        module.execute_module()
    except Exception as e:
        module.fail_json(msg=str(e), exception=traceback.format_exc())


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: mso_schema_site_anp_epg_staticport
short_description: Manage site-local EPG static ports in schema template
description:
- Manage site-local EPG static ports in schema template on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: yes
  site:
    description:
    - The name of the site.
    type: str
    required: yes
  template:
    description:
    - The name of the template.
    type: str
    required: yes
  anp:
    description:
    - The name of the ANP.
    type: str
    required: yes
  epg:
    description:
    - The name of the EPG.
    type: str
    required: yes
  type:
    description:
    - The path type of the static port
    - vpc is used for a Virtual Port Channel
    - dpc is used for a Direct Port Channel
    - port is used for a single interface
    type: str
    choices: [ port, vpc, dpc ]
    default: port
  pod:
    description:
    - The pod of the static port.
    type: str
  leaf:
    description:
    - The leaf of the static port.
    type: str
  fex:
    description:
    - The fex id of the static port.
    type: str
  path:
    description:
    - The path of the static port.
    type: str
  vlan:
    description:
    - The port encap VLAN id of the static port.
    type: int
  deployment_immediacy:
    description:
    - The deployment immediacy of the static port.
    - C(immediate) means B(Deploy immediate).
    - C(lazy) means B(deploy on demand).
    type: str
    choices: [ immediate, lazy ]
    default: lazy
  mode:
    description:
    - The mode of the static port.
    - C(native) means B(Access (802.1p)).
    - C(regular) means B(Trunk).
    - C(untagged) means B(Access (untagged)).
    type: str
    choices: [ native, regular, untagged ]
    default: untagged
  primary_micro_segment_vlan:
    description:
    - Primary micro-seg VLAN of static port.
    type: int
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- The ACI MultiSite PATCH API has a deficiency requiring some objects to be referenced by index.
  This can cause silent corruption on concurrent access when changing/removing an object as
  the wrong object may be referenced. This module is affected by this deficiency.
seealso:
- module: cisco.mso.mso_schema_site_anp_epg
- module: cisco.mso.mso_schema_template_anp_epg
extends_documentation_fragment: cisco.mso.modules
'''

EXAMPLES = r'''
- name: Add a new static port to a site EPG
  cisco.mso.mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    type: port
    pod: pod-1
    leaf: 101
    path: eth1/1
    vlan: 126
    deployment_immediacy: immediate
    state: present
  delegate_to: localhost

- name: Add a new static fex port to a site EPG
  mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    type: port
    pod: pod-1
    leaf: 101
    fex: 151
    path: eth1/1
    vlan: 126
    deployment_immediacy: lazy
    state: present
  delegate_to: localhost

- name: Add a new static VPC to a site EPG
  mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    pod: pod-1
    leaf: 101-102
    path: ansible_polgrp
    vlan: 127
    type: vpc
    mode: untagged
    deployment_immediacy: lazy
    state: present
  delegate_to: localhost

- name: Remove a static port from a site EPG
  cisco.mso.mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    type: port
    pod: pod-1
    leaf: 101
    path: eth1/1
    state: absent
  delegate_to: localhost

- name: Query a specific site EPG static port
  cisco.mso.mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    type: port
    pod: pod-1
    leaf: 101
    path: eth1/1
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all site EPG static ports
  cisco.mso.mso_schema_site_anp_epg_staticport:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    state: query
  delegate_to: localhost
  register: query_result
'''

RETURN = r'''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type='str', required=True),
        site=dict(type='str', required=True),
        template=dict(type='str', required=True),
        anp=dict(type='str', required=True),
        epg=dict(type='str', required=True),
        type=dict(type='str', default='port', choices=['port', 'vpc', 'dpc']),
        pod=dict(type='str'),  # This parameter is not required for querying all objects
        leaf=dict(type='str'),  # This parameter is not required for querying all objects
        fex=dict(type='str'),    # This parameter is not required for querying all objects
        path=dict(type='str'),  # This parameter is not required for querying all objects
        vlan=dict(type='int'),  # This parameter is not required for querying all objects
        primary_micro_segment_vlan=dict(type='int'),  # This parameter is not required for querying all objects
        deployment_immediacy=dict(type='str', default='lazy', choices=['immediate', 'lazy']),
        mode=dict(type='str', default='untagged', choices=['native', 'regular', 'untagged']),
        state=dict(type='str', default='present', choices=['absent', 'present', 'query']),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ['state', 'absent', ['type', 'pod', 'leaf', 'path', 'vlan']],
            ['state', 'present', ['type', 'pod', 'leaf', 'path', 'vlan']],
        ],
    )

    schema = module.params.get('schema')
    site = module.params.get('site')
    template = module.params.get('template').replace(' ', '')
    anp = module.params.get('anp')
    epg = module.params.get('epg')
    path_type = module.params.get('type')
    pod = module.params.get('pod')
    leaf = module.params.get('leaf')
    fex = module.params.get('fex')
    path = module.params.get('path')
    vlan = module.params.get('vlan')
    primary_micro_segment_vlan = module.params.get('primary_micro_segment_vlan')
    deployment_immediacy = module.params.get('deployment_immediacy')
    mode = module.params.get('mode')
    state = module.params.get('state')

    if path_type == 'port' and fex is not None:
        # Select port path for fex if fex param is used
        portpath = 'topology/{0}/paths-{1}/extpaths-{2}/pathep-[{3}]'.format(pod, leaf, fex, path)
    elif path_type == 'vpc':
        portpath = 'topology/{0}/protpaths-{1}/pathep-[{2}]'.format(pod, leaf, path)
    else:
        portpath = 'topology/{0}/paths-{1}/pathep-[{2}]'.format(pod, leaf, path)

    mso = MSOModule(module)

    # Get schema_id
    schema_obj = mso.get_obj('schemas', displayName=schema)
    if not schema_obj:
        mso.fail_json(msg="Provided schema '{0}' does not exist".format(schema))
    schema_path = 'schemas/{id}'.format(**schema_obj)
    schema_id = schema_obj.get('id')

    # Get template
    templates = [t.get('name') for t in schema_obj.get('templates')]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ', '.join(templates)))
    template_idx = templates.index(template)

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if 'sites' not in schema_obj:
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get('siteId'), s.get('templateName')) for s in schema_obj.get('sites')]
    sites_list = [s.get('siteId') + '/' + s.get('templateName') for s in schema_obj.get('sites')]
    if (site_id, template) not in sites:
        mso.fail_json(msg="Provided site/siteId/template '{0}/{1}/{2}' does not exist. "
                          "Existing siteIds/templates: {3}".format(site, site_id, template, ', '.join(sites_list)))

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = '{0}-{1}'.format(site_id, template)

    payload = dict()
    ops = []
    op_path = ''

    # Get ANP
    anp_ref = mso.anp_ref(schema_id=schema_id, template=template, anp=anp)
    anps = [a.get('anpRef') for a in schema_obj['sites'][site_idx]['anps']]
    anps_in_temp = [a.get('name') for a in schema_obj['templates'][template_idx]['anps']]
    if anp not in anps_in_temp:
        mso.fail_json(msg="Provided anp '{0}' does not exist. Existing anps: {1}".format(anp, ', '.join(anps)))
    else:
        # Update anp index at template level
        template_anp_idx = anps_in_temp.index(anp)

    # If anp not at site level but exists at template level
    if anp_ref not in anps:
        op_path = '/sites/{0}/anps/-'.format(site_template)
        payload.update(
            anpRef=dict(
                schemaId=schema_id,
                templateName=template,
                anpName=anp,
            ),
        )

    else:
        # Update anp index at site level
        anp_idx = anps.index(anp_ref)

    # Get EPG
    epg_ref = mso.epg_ref(schema_id=schema_id, template=template, anp=anp, epg=epg)

    # If anp exists at site level
    if 'anpRef' not in payload:
        epgs = [e.get('epgRef') for e in schema_obj['sites'][site_idx]['anps'][anp_idx]['epgs']]

    # If anp already at site level AND if epg not at site level (or) anp not at site level
    if ('anpRef' not in payload and epg_ref not in epgs) or 'anpRef' in payload:
        epgs_in_temp = [e.get('name') for e in schema_obj['templates'][template_idx]['anps'][template_anp_idx]['epgs']]

        # If EPG not at template level - Fail
        if epg not in epgs_in_temp:
            mso.fail_json(msg="Provided EPG '{0}' does not exist. Existing EPGs: {1} epgref {2}".format(epg, ', '.join(epgs_in_temp), epg_ref))

        # EPG at template level but not at site level. Create payload at site level for EPG
        else:

            new_epg = dict(
                epgRef=dict(
                    schemaId=schema_id,
                    templateName=template,
                    anpName=anp,
                    epgName=epg,
                )
            )

            # If anp not in payload then, anp already exists at site level. New payload will only have new EPG payload
            if 'anpRef' not in payload:
                op_path = '/sites/{0}/anps/{1}/epgs/-'.format(site_template, anp)
                payload = new_epg
            else:
                # If anp in payload, anp exists at site level. Update payload with EPG payload
                payload['epgs'] = [new_epg]

    # Update index of EPG at site level
    else:
        epg_idx = epgs.index(epg_ref)

    # Get Leaf
    # If anp at site level and epg is at site level
    if 'anpRef' not in payload and 'epgRef' not in payload:
        portpaths = [p.get('path') for p in schema_obj.get('sites')[site_idx]['anps'][anp_idx]['epgs'][epg_idx]['staticPorts']]
        if portpath in portpaths:
            portpath_idx = portpaths.index(portpath)
            port_path = '/sites/{0}/anps/{1}/epgs/{2}/staticPorts/{3}'.format(site_template, anp, epg, portpath_idx)
            mso.existing = schema_obj.get('sites')[site_idx]['anps'][anp_idx]['epgs'][epg_idx]['staticPorts'][portpath_idx]

    if state == 'query':
        if leaf is None or vlan is None:
            mso.existing = schema_obj.get('sites')[site_idx]['anps'][anp_idx]['epgs'][epg_idx]['staticPorts']
        elif not mso.existing:
            mso.fail_json(msg="Static port '{portpath}' not found".format(portpath=portpath))
        mso.exit_json()

    ports_path = '/sites/{0}/anps/{1}/epgs/{2}/staticPorts'.format(site_template, anp, epg)
    ops = []
    new_leaf = dict(
        deploymentImmediacy=deployment_immediacy,
        mode=mode,
        path=portpath,
        portEncapVlan=vlan,
        type=path_type,
        microSegVlan=primary_micro_segment_vlan,
    )

    # If payload is empty, anp and EPG already exist at site level
    if not payload:
        op_path = ports_path + '/-'
        payload = new_leaf

    # If payload exists
    else:
        # If anp already exists at site level
        if 'anpRef' not in payload:
            payload['staticPorts'] = [new_leaf]
        else:
            payload['epgs'][0]['staticPorts'] = [new_leaf]

    mso.previous = mso.existing
    if state == 'absent':
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op='remove', path=port_path))

    elif state == 'present':

        mso.sanitize(payload, collate=True)

        if mso.existing:
            ops.append(dict(op='replace', path=port_path, value=mso.sent))
        else:
            ops.append(dict(op='add', path=op_path, value=mso.sent))

        mso.existing = new_leaf

    if not module.check_mode:
        mso.request(schema_path, method='PATCH', data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

__metaclass__ = type

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = """
module: nxos_pim_interface
extends_documentation_fragment:
- cisco.nxos.nxos
short_description: Manages PIM interface configuration.
description:
- Manages PIM interface configuration settings.
version_added: 1.0.0
author:
- Jason Edelman (@jedelman8)
notes:
- Tested against NXOSv 7.3.(0)D1(1) on VIRL
- When C(state=default), supported params will be reset to a default state. These
  include C(dr_prio), C(hello_auth_key), C(hello_interval), C(jp_policy_out), C(jp_policy_in),
  C(jp_type_in), C(jp_type_out), C(border), C(neighbor_policy), C(neighbor_type).
- The C(hello_auth_key) param is not idempotent.
- C(hello_auth_key) only supports clear text passwords.
- When C(state=absent), pim interface configuration will be set to defaults and pim-sm
  will be disabled on the interface.
- PIM must be enabled on the device to use this module.
- This module is for Layer 3 interfaces.
options:
  interface:
    description:
    - Full name of the interface such as Ethernet1/33.
    type: str
    required: true
  sparse:
    description:
    - Enable/disable sparse-mode on the interface.
    type: bool
    default: false
  bfd:
    description:
    - Enables BFD for PIM at the interface level. This overrides the bfd variable
      set at the pim global level.
    - Valid values are 'enable', 'disable' or 'default'.
    - "Dependency: ''feature bfd''"
    type: str
    choices:
    - enable
    - disable
    - default
  dr_prio:
    description:
    - Configures priority for PIM DR election on interface.
    type: str
  hello_auth_key:
    description:
    - Authentication for hellos on this interface.
    type: str
  hello_interval:
    description:
    - Hello interval in milliseconds for this interface.
    type: int
  jp_policy_out:
    description:
    - Policy for join-prune messages (outbound).
    type: str
  jp_policy_in:
    description:
    - Policy for join-prune messages (inbound).
    type: str
  jp_type_out:
    description:
    - Type of policy mapped to C(jp_policy_out).
    type: str
    choices:
    - prefix
    - routemap
  jp_type_in:
    description:
    - Type of policy mapped to C(jp_policy_in).
    type: str
    choices:
    - prefix
    - routemap
  border:
    description:
    - Configures interface to be a boundary of a PIM domain.
    type: bool
    default: false
  neighbor_policy:
    description:
    - Configures a neighbor policy for filtering adjacencies.
    type: str
  neighbor_type:
    description:
    - Type of policy mapped to neighbor_policy.
    type: str
    choices:
    - prefix
    - routemap
  state:
    description:
    - Manages desired state of the resource.
    type: str
    choices:
    - present
    - absent
    - default
    default: present
"""
EXAMPLES = """
- name: Ensure PIM is not running on the interface
  cisco.nxos.nxos_pim_interface:
    interface: eth1/33
    state: absent

- name: Ensure the interface has pim-sm enabled with the appropriate priority and
    hello interval
  cisco.nxos.nxos_pim_interface:
    interface: eth1/33
    dr_prio: 10
    hello_interval: 40
    state: present

- name: Ensure join-prune policies exist
  cisco.nxos.nxos_pim_interface:
    interface: eth1/33
    jp_policy_in: JPIN
    jp_policy_out: JPOUT
    jp_type_in: routemap
    jp_type_out: routemap

- name: disable bfd on the interface
  cisco.nxos.nxos_pim_interface:
    interface: eth1/33
    bfd: disable

- name: Ensure defaults are in place
  cisco.nxos.nxos_pim_interface:
    interface: eth1/33
    state: default
"""

RETURN = r"""
commands:
    description: command sent to the device
    returned: always
    type: list
    sample: ["interface eth1/33",
             "ip pim neighbor-policy test",
             "ip pim bfd-instance disable",
             "ip pim neighbor-policy test"
            ]
"""

import re
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    get_config,
    load_config,
    run_commands,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    nxos_argument_spec,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    get_interface_type,
)


PARAM_TO_COMMAND_KEYMAP = {
    "interface": "",
    "bfd": "ip pim bfd-instance",
    "sparse": "ip pim sparse-mode",
    "dr_prio": "ip pim dr-priority {0}",
    "hello_interval": "ip pim hello-interval {0}",
    "hello_auth_key": "ip pim hello-authentication ah-md5 {0}",
    "border": "ip pim border",
    "jp_policy_out": "ip pim jp-policy prefix-list {0} out",
    "jp_policy_in": "ip pim jp-policy prefix-list {0} in",
    "jp_type_in": "",
    "jp_type_out": "",
    "neighbor_policy": "ip pim neighbor-policy prefix-list {0}",
    "neighbor_type": "",
}

PARAM_TO_DEFAULT_KEYMAP = {
    "bfd": "default",
    "dr_prio": "1",
    "hello_interval": "30000",
    "sparse": False,
    "border": False,
    "hello_auth_key": False,
}

BFD_KEYMAP = {
    None: None,
    "default": "no ip pim bfd-instance",
    "disable": "ip pim bfd-instance disable",
    "enable": "ip pim bfd-instance",
}


def execute_show_command(command, module, text=False):
    if text:
        cmds = [{"command": command, "output": "text"}]
    else:
        cmds = [{"command": command, "output": "json"}]

    return run_commands(module, cmds)


def flatten_list(command_lists):
    flat_command_list = []
    for command in command_lists:
        if isinstance(command, list):
            flat_command_list.extend(command)
        else:
            flat_command_list.append(command)
    return flat_command_list


def local_existing(gexisting):
    jp_bidir = False
    isauth = False
    if gexisting:
        jp_bidir = gexisting.get("jp_bidir")
        isauth = gexisting.get("isauth")
        if jp_bidir and isauth:
            gexisting.pop("jp_bidir")
            gexisting.pop("isauth")

    return gexisting, jp_bidir, isauth


def get_interface_mode(interface, intf_type, module):
    mode = "unknown"
    command = "show interface {0}".format(interface)
    body = execute_show_command(command, module)

    try:
        interface_table = body[0]["TABLE_interface"]["ROW_interface"]
    except (KeyError, AttributeError, IndexError):
        return mode

    if intf_type in ["ethernet", "portchannel"]:
        mode = str(interface_table.get("eth_mode", "layer3"))
        if mode in ["access", "trunk"]:
            mode = "layer2"
        elif mode == "routed":
            mode = "layer3"
    elif intf_type in ["loopback", "svi"]:
        mode = "layer3"
    return mode


def get_pim_interface(module, interface):
    pim_interface = {}
    body = get_config(module, flags=["interface {0}".format(interface)])

    pim_interface["bfd"] = "default"
    pim_interface["neighbor_type"] = None
    pim_interface["neighbor_policy"] = None
    pim_interface["jp_policy_in"] = None
    pim_interface["jp_policy_out"] = None
    pim_interface["jp_type_in"] = None
    pim_interface["jp_type_out"] = None
    pim_interface["jp_bidir"] = False
    pim_interface["isauth"] = False

    if body:
        all_lines = body.splitlines()

        for each in all_lines:
            if "jp-policy" in each:
                policy_name = re.search(
                    r"ip pim jp-policy(?: prefix-list)? (\S+)(?: \S+)?", each
                ).group(1)
                if "prefix-list" in each:
                    ptype = "prefix"
                else:
                    ptype = "routemap"
                if "out" in each:
                    pim_interface["jp_policy_out"] = policy_name
                    pim_interface["jp_type_out"] = ptype
                elif "in" in each:
                    pim_interface["jp_policy_in"] = policy_name
                    pim_interface["jp_type_in"] = ptype
                else:
                    pim_interface["jp_policy_in"] = policy_name
                    pim_interface["jp_policy_out"] = policy_name
                    pim_interface["jp_bidir"] = True
            elif "neighbor-policy" in each:
                pim_interface["neighbor_policy"] = re.search(
                    r"ip pim neighbor-policy(?: prefix-list)? (\S+)", each
                ).group(1)
                if "prefix-list" in each:
                    pim_interface["neighbor_type"] = "prefix"
                else:
                    pim_interface["neighbor_type"] = "routemap"
            elif "ah-md5" in each:
                pim_interface["isauth"] = True
            elif "sparse-mode" in each:
                pim_interface["sparse"] = True
            elif "bfd-instance" in each:
                m = re.search(
                    r"ip pim bfd-instance(?P<disable> disable)?", each
                )
                if m:
                    pim_interface["bfd"] = (
                        "disable" if m.group("disable") else "enable"
                    )
            elif "border" in each:
                pim_interface["border"] = True
            elif "hello-interval" in each:
                pim_interface["hello_interval"] = re.search(
                    r"ip pim hello-interval (\d+)", body
                ).group(1)
            elif "dr-priority" in each:
                pim_interface["dr_prio"] = re.search(
                    r"ip pim dr-priority (\d+)", body
                ).group(1)

    return pim_interface


def fix_delta(delta, existing):
    for key in list(delta):
        if key in ["dr_prio", "hello_interval", "sparse", "border"]:
            if (
                delta.get(key) == PARAM_TO_DEFAULT_KEYMAP.get(key)
                and existing.get(key) is None
            ):
                delta.pop(key)
    return delta


def config_pim_interface(delta, existing, jp_bidir, isauth):
    command = None
    commands = []

    delta = fix_delta(delta, existing)

    if jp_bidir:
        if delta.get("jp_policy_in") or delta.get("jp_policy_out"):
            if existing.get("jp_type_in") == "prefix":
                command = "no ip pim jp-policy prefix-list {0}".format(
                    existing.get("jp_policy_in")
                )
            else:
                command = "no ip pim jp-policy {0}".format(
                    existing.get("jp_policy_in")
                )
            if command:
                commands.append(command)

    for k, v in delta.items():
        if k in [
            "bfd",
            "dr_prio",
            "hello_interval",
            "hello_auth_key",
            "border",
            "sparse",
        ]:
            if k == "bfd":
                command = BFD_KEYMAP[v]
            elif v:
                command = PARAM_TO_COMMAND_KEYMAP.get(k).format(v)
            elif k == "hello_auth_key":
                if isauth:
                    command = "no ip pim hello-authentication ah-md5"
            else:
                command = "no " + PARAM_TO_COMMAND_KEYMAP.get(k).format(v)

            if command:
                commands.append(command)
        elif k in [
            "neighbor_policy",
            "jp_policy_in",
            "jp_policy_out",
            "neighbor_type",
        ]:
            if k in ["neighbor_policy", "neighbor_type"]:
                temp = delta.get("neighbor_policy") or existing.get(
                    "neighbor_policy"
                )
                if delta.get("neighbor_type") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif delta.get("neighbor_type") == "routemap":
                    command = "ip pim neighbor-policy {0}".format(temp)
                elif existing.get("neighbor_type") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif existing.get("neighbor_type") == "routemap":
                    command = "ip pim neighbor-policy {0}".format(temp)
            elif k in ["jp_policy_in", "jp_type_in"]:
                temp = delta.get("jp_policy_in") or existing.get(
                    "jp_policy_in"
                )
                if delta.get("jp_type_in") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif delta.get("jp_type_in") == "routemap":
                    command = "ip pim jp-policy {0} in".format(temp)
                elif existing.get("jp_type_in") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif existing.get("jp_type_in") == "routemap":
                    command = "ip pim jp-policy {0} in".format(temp)
            elif k in ["jp_policy_out", "jp_type_out"]:
                temp = delta.get("jp_policy_out") or existing.get(
                    "jp_policy_out"
                )
                if delta.get("jp_type_out") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif delta.get("jp_type_out") == "routemap":
                    command = "ip pim jp-policy {0} out".format(temp)
                elif existing.get("jp_type_out") == "prefix":
                    command = PARAM_TO_COMMAND_KEYMAP.get(k).format(temp)
                elif existing.get("jp_type_out") == "routemap":
                    command = "ip pim jp-policy {0} out".format(temp)
            if command:
                commands.append(command)
        command = None

    if "no ip pim sparse-mode" in commands:
        # sparse is long-running on some platforms, process it last
        commands.remove("no ip pim sparse-mode")
        commands.append("no ip pim sparse-mode")
    return commands


def get_pim_interface_defaults():

    args = dict(
        dr_prio=PARAM_TO_DEFAULT_KEYMAP.get("dr_prio"),
        bfd=PARAM_TO_DEFAULT_KEYMAP.get("bfd"),
        border=PARAM_TO_DEFAULT_KEYMAP.get("border"),
        sparse=PARAM_TO_DEFAULT_KEYMAP.get("sparse"),
        hello_interval=PARAM_TO_DEFAULT_KEYMAP.get("hello_interval"),
        hello_auth_key=PARAM_TO_DEFAULT_KEYMAP.get("hello_auth_key"),
    )

    default = dict(
        (param, value) for (param, value) in args.items() if value is not None
    )

    return default


def default_pim_interface_policies(existing, jp_bidir):
    commands = []

    if jp_bidir:
        if existing.get("jp_policy_in") or existing.get("jp_policy_out"):
            if existing.get("jp_type_in") == "prefix":
                command = "no ip pim jp-policy prefix-list {0}".format(
                    existing.get("jp_policy_in")
                )
        if command:
            commands.append(command)

    elif not jp_bidir:
        command = None
        for k in existing:
            if k == "jp_policy_in":
                if existing.get("jp_policy_in"):
                    if existing.get("jp_type_in") == "prefix":
                        command = "no ip pim jp-policy prefix-list {0} in".format(
                            existing.get("jp_policy_in")
                        )
                    else:
                        command = "no ip pim jp-policy {0} in".format(
                            existing.get("jp_policy_in")
                        )
            elif k == "jp_policy_out":
                if existing.get("jp_policy_out"):
                    if existing.get("jp_type_out") == "prefix":
                        command = "no ip pim jp-policy prefix-list {0} out".format(
                            existing.get("jp_policy_out")
                        )
                    else:
                        command = "no ip pim jp-policy {0} out".format(
                            existing.get("jp_policy_out")
                        )
            if command:
                commands.append(command)
            command = None

    if existing.get("neighbor_policy"):
        command = "no ip pim neighbor-policy"
        commands.append(command)

    return commands


def config_pim_interface_defaults(existing, jp_bidir, isauth):
    command = []

    # returns a dict
    defaults = get_pim_interface_defaults()
    delta = dict(set(defaults.items()).difference(existing.items()))
    if delta:
        # returns a list
        command = config_pim_interface(delta, existing, jp_bidir, isauth)
    comm = default_pim_interface_policies(existing, jp_bidir)
    if comm:
        for each in comm:
            command.append(each)

    return command


def normalize_proposed_values(proposed):
    keys = proposed.keys()
    if "bfd" in keys:
        # bfd is a tri-state string: enable, disable, default
        proposed["bfd"] = proposed["bfd"].lower()
    if "hello_interval" in keys:
        proposed["hello_interval"] = str(proposed["hello_interval"] * 1000)


def main():
    argument_spec = dict(
        interface=dict(type="str", required=True),
        sparse=dict(type="bool", default=False),
        dr_prio=dict(type="str"),
        hello_auth_key=dict(type="str", no_log=True),
        hello_interval=dict(type="int"),
        jp_policy_out=dict(type="str"),
        jp_policy_in=dict(type="str"),
        jp_type_out=dict(type="str", choices=["prefix", "routemap"]),
        jp_type_in=dict(type="str", choices=["prefix", "routemap"]),
        bfd=dict(type="str", choices=["enable", "disable", "default"]),
        border=dict(type="bool", default=False),
        neighbor_policy=dict(type="str"),
        neighbor_type=dict(type="str", choices=["prefix", "routemap"]),
        state=dict(
            type="str",
            default="present",
            choices=["absent", "default", "present"],
        ),
    )
    argument_spec.update(nxos_argument_spec)

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True
    )

    warnings = list()
    results = {"changed": False, "commands": [], "warnings": warnings}

    state = module.params["state"]
    interface = module.params["interface"]
    jp_type_in = module.params["jp_type_in"]
    jp_type_out = module.params["jp_type_out"]
    jp_policy_in = module.params["jp_policy_in"]
    jp_policy_out = module.params["jp_policy_out"]
    neighbor_policy = module.params["neighbor_policy"]
    neighbor_type = module.params["neighbor_type"]

    intf_type = get_interface_type(interface)
    if get_interface_mode(interface, intf_type, module) == "layer2":
        module.fail_json(msg="this module only works on Layer 3 interfaces.")

    if jp_policy_in:
        if not jp_type_in:
            module.fail_json(
                msg="jp_type_in required when using jp_policy_in."
            )
    if jp_policy_out:
        if not jp_type_out:
            module.fail_json(
                msg="jp_type_out required when using jp_policy_out."
            )
    if neighbor_policy:
        if not neighbor_type:
            module.fail_json(
                msg="neighbor_type required when using neighbor_policy."
            )

    get_existing = get_pim_interface(module, interface)
    existing, jp_bidir, isauth = local_existing(get_existing)

    args = PARAM_TO_COMMAND_KEYMAP.keys()
    proposed = dict(
        (k, v) for k, v in module.params.items() if v is not None and k in args
    )
    normalize_proposed_values(proposed)

    delta = dict(set(proposed.items()).difference(existing.items()))

    commands = []
    if state == "present":
        if delta:
            command = config_pim_interface(delta, existing, jp_bidir, isauth)
            if command:
                commands.append(command)
    elif state == "default" or state == "absent":
        defaults = config_pim_interface_defaults(existing, jp_bidir, isauth)
        if defaults:
            commands.append(defaults)

    if commands:
        commands.insert(0, ["interface {0}".format(interface)])

    cmds = flatten_list(commands)
    if cmds:
        results["changed"] = True
        if not module.check_mode:
            load_config(module, cmds)
        if "configure" in cmds:
            cmds.pop(0)

    results["commands"] = cmds

    module.exit_json(**results)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The nxos_acls class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import socket
import re
from copy import deepcopy
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    to_list,
    remove_empties,
    dict_diff,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import (
    Facts,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.acls.acls import (
    AclsArgs,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import (
    utils,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    flatten_dict,
    search_obj_in_list,
    get_interface_type,
    normalize_interface,
)


class Acls(ConfigBase):
    """
    The nxos_acls class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["acls"]

    def __init__(self, module):
        super(Acls, self).__init__(module)

    def get_acls_facts(self, data=None):
        """ Get the 'facts' (the current configuration)

        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset, self.gather_network_resources, data=data
        )
        acls_facts = facts["ansible_network_resources"].get("acls")
        if not acls_facts:
            return []
        return acls_facts

    def edit_config(self, commands):
        """Wrapper method for `_connection.edit_config()`
        This exists solely to allow the unit test framework to mock device connection calls.
        """
        return self._connection.edit_config(commands)

    def execute_module(self):
        """ Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        warnings = list()
        commands = list()
        state = self._module.params["state"]
        action_states = ["merged", "replaced", "deleted", "overridden"]

        if state == "gathered":
            result["gathered"] = self.get_acls_facts()
        elif state == "rendered":
            result["rendered"] = self.set_config({})
        elif state == "parsed":
            result["parsed"] = self.set_config({})
        else:
            existing_acls_facts = self.get_acls_facts()
            commands.extend(self.set_config(existing_acls_facts))
            if commands and state in action_states:
                if not self._module.check_mode:
                    self._connection.edit_config(commands)
                result["changed"] = True
                result["before"] = existing_acls_facts
                result["commands"] = commands

            changed_acls_facts = self.get_acls_facts()
            if result["changed"]:
                result["after"] = changed_acls_facts
        result["warnings"] = warnings
        return result

    def set_config(self, existing_acls_facts):
        """ Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        config = self._module.params["config"]
        want = []
        if config:
            for w in config:
                want.append(remove_empties(w))
        have = existing_acls_facts
        if want:
            want = self.convert_values(want)
        resp = self.set_state(want, have)
        return to_list(resp)

    def convert_values(self, want):
        """
        This method is used to map and convert the user given values with what will actually be present in the device configuation
        """
        port_protocol = {
            515: "lpd",
            517: "talk",
            7: "echo",
            9: "discard",
            12: "exec",
            13: "login",
            14: "cmd",
            109: "pop2",
            19: "chargen",
            20: "ftp-data",
            21: "ftp",
            23: "telnet",
            25: "smtp",
            540: "uucp",
            543: "klogin",
            544: "kshell",
            37: "time",
            43: "whois",
            49: "tacacs",
            179: "bgp",
            53: "domain",
            194: "irc",
            70: "gopher",
            79: "finger",
            80: "www",
            101: "hostname",
            3949: "drip",
            110: "pop3",
            111: "sunrpc",
            496: "pim-auto-rp",
            113: "ident",
            119: "nntp",
        }
        protocol = {
            1: "icmp",
            2: "igmp",
            4: "ip",
            6: "tcp",
            103: "pim",
            108: "pcp",
            47: "gre",
            17: "udp",
            50: "esp",
            51: "ahp",
            88: "eigrp",
            89: "ospf",
            94: "nos",
        }
        precedence = {
            0: "routine",
            1: "priority",
            2: "immediate",
            3: "flash",
            4: "flash-override",
            5: "critical",
            6: "internet",
            7: "network",
        }
        dscp = {
            10: "AF11",
            12: "AF12",
            14: "AF13",
            18: "AF21",
            20: "AF22",
            22: "AF23",
            26: "AF31",
            28: "AF32",
            30: "AF33",
            34: "AF41",
            36: "AF42",
            38: "AF43",
            8: "CS1",
            16: "CS2",
            24: "CS3",
            32: "CS4",
            40: "CS5",
            48: "CS6",
            56: "CS7",
            0: "Default",
            46: "EF",
        }
        # port_pro_num = list(protocol.keys())
        for afi in want:
            if "acls" in afi.keys():
                for acl in afi["acls"]:
                    if "aces" in acl.keys():
                        for ace in acl["aces"]:
                            if "dscp" in ace.keys():
                                if ace["dscp"].isdigit():
                                    ace["dscp"] = dscp[int(ace["dscp"])]
                                ace["dscp"] = ace["dscp"].lower()
                            if "precedence" in ace.keys():
                                if ace["precedence"].isdigit():
                                    ace["precedence"] = precedence[
                                        int(ace["precedence"])
                                    ]
                            if (
                                "protocol" in ace.keys()
                                and ace["protocol"].isdigit()
                                and int(ace["protocol"]) in protocol.keys()
                            ):
                                ace["protocol"] = protocol[
                                    int(ace["protocol"])
                                ]
                                # convert number to name
                            if "protocol" in ace.keys() and ace[
                                "protocol"
                            ] in ["tcp", "udp"]:
                                for end in ["source", "destination"]:
                                    if "port_protocol" in ace[end].keys():
                                        key = list(
                                            ace[end]["port_protocol"].keys()
                                        )[0]
                                        # key could be eq,gt,lt,neq or range
                                        if key != "range":
                                            val = ace[end]["port_protocol"][
                                                key
                                            ]
                                            if (
                                                val.isdigit()
                                                and int(val)
                                                in port_protocol.keys()
                                            ):
                                                ace[end]["port_protocol"][
                                                    key
                                                ] = port_protocol[int(val)]
                                        else:
                                            st = int(
                                                ace[end]["port_protocol"][
                                                    "range"
                                                ]["start"]
                                            )

                                            end = int(
                                                ace[end]["port_protocol"][
                                                    "range"
                                                ]["end"]
                                            )

                                            if st in port_protocol.keys():
                                                ace[end]["port_protocol"][
                                                    "range"
                                                ]["start"] = port_protocol[st]
                                            if end in port_protocol.keys():
                                                ace[end]["port_protocol"][
                                                    "range"
                                                ]["end"] = port_protocol[end]
        return want

    def set_state(self, want, have):
        """ Select the appropriate function based on the state provided

        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        state = self._module.params["state"]
        commands = []
        if state == "overridden":
            commands = self._state_overridden(want, have)
        elif state == "deleted":
            commands = self._state_deleted(want, have)
        elif state == "rendered":
            commands = self._state_rendered(want)
        elif state == "parsed":
            want = self._module.params["running_config"]
            commands = self._state_parsed(want)
        else:
            for w in want:
                if state == "merged":
                    commands.extend(self._state_merged(w, have))
                elif state == "replaced":
                    commands.extend(self._state_replaced(w, have))
        if state != "parsed":
            commands = [c.strip() for c in commands]
        return commands

    def _state_parsed(self, want):
        return self.get_acls_facts(want)

    def _state_rendered(self, want):
        commands = []
        for w in want:
            commands.extend(self.set_commands(w, {}))
        return commands

    def _state_replaced(self, want, have):
        """ The command generator when state is replaced

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        have_afi = search_obj_in_list(want["afi"], have, "afi")
        del_dict = {"acls": []}
        want_names = []
        if have_afi != want:
            if have_afi:
                del_dict.update({"afi": have_afi["afi"], "acls": []})
                if want.get("acls"):
                    want_names = [w["name"] for w in want["acls"]]
                    have_names = [h["name"] for h in have_afi["acls"]]
                    want_acls = want.get("acls")
                    for w in want_acls:
                        acl_commands = []
                        if w["name"] not in have_names:
                            # creates new ACL in replaced state
                            merge_dict = {"afi": want["afi"], "acls": [w]}
                            commands.extend(
                                self._state_merged(merge_dict, have)
                            )
                        else:
                            # acl in want exists in have
                            have_name = search_obj_in_list(
                                w["name"], have_afi["acls"], "name"
                            )
                            have_aces = (
                                have_name.get("aces")
                                if have_name.get("aces")
                                else []
                            )
                            merge_aces = []
                            del_aces = []
                            w_aces = w.get("aces") if w.get("aces") else []

                            for ace in have_aces:
                                if ace not in w_aces:
                                    del_aces.append(ace)
                            for ace in w_aces:
                                if ace not in have_aces:
                                    merge_aces.append(ace)
                            merge_dict = {
                                "afi": want["afi"],
                                "acls": [
                                    {"name": w["name"], "aces": merge_aces}
                                ],
                            }
                            del_dict = {
                                "afi": want["afi"],
                                "acls": [
                                    {"name": w["name"], "aces": del_aces}
                                ],
                            }
                            if del_dict["acls"]:
                                acl_commands.extend(
                                    self._state_deleted([del_dict], have)
                                )
                            acl_commands.extend(
                                self._state_merged(merge_dict, have)
                            )

                            for i in range(1, len(acl_commands)):
                                if acl_commands[i] == acl_commands[0]:
                                    acl_commands[i] = ""
                            commands.extend(acl_commands)
                else:
                    acls = []
                    # no acls given in want, so delete all have acls
                    for acl in have_afi["acls"]:
                        acls.append({"name": acl["name"]})
                    del_dict["acls"] = acls
                    if del_dict["acls"]:
                        commands.extend(self._state_deleted([del_dict], have))

            else:
                # want_afi is not present in have
                commands.extend(self._state_merged(want, have))

        commands = list(filter(None, commands))
        return commands

    def _state_overridden(self, want, have):
        """ The command generator when state is overridden

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        want_afi = [w["afi"] for w in want]
        for h in have:
            if h["afi"] in want_afi:
                w = search_obj_in_list(h["afi"], want, "afi")
                for h_acl in h["acls"]:
                    w_acl = search_obj_in_list(
                        h_acl["name"], w["acls"], "name"
                    )
                    if not w_acl:
                        del_dict = {
                            "afi": h["afi"],
                            "acls": [{"name": h_acl["name"]}],
                        }
                        commands.extend(self._state_deleted([del_dict], have))
            else:
                # if afi is not in want
                commands.extend(self._state_deleted([{"afi": h["afi"]}], have))
        for w in want:
            commands.extend(self._state_replaced(w, have))
        return commands

    def _state_merged(self, want, have):
        """ The command generator when state is merged

        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        return self.set_commands(want, have)

    def _state_deleted(self, want, have):
        """ The command generator when state is deleted

        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commands = []
        if want:  # and have != want:
            for w in want:
                ip = "ipv6" if w["afi"] == "ipv6" else "ip"
                acl_names = []
                have_afi = search_obj_in_list(w["afi"], have, "afi")
                # if want['afi] not in have, ignore
                if have_afi:
                    if w.get("acls"):
                        for acl in w["acls"]:
                            if (
                                "aces" in acl.keys()
                                and self._module.params["state"] != "deleted"
                            ):
                                have_name = search_obj_in_list(
                                    acl["name"], have_afi["acls"], "name"
                                )
                                if have_name:
                                    ace_commands = []
                                    flag = 0
                                    for ace in acl["aces"]:
                                        if list(ace.keys()) == ["sequence"]:
                                            # only sequence number is specified to be deleted
                                            if "aces" in have_name.keys():
                                                for h_ace in have_name["aces"]:
                                                    if (
                                                        h_ace["sequence"]
                                                        == ace["sequence"]
                                                    ):
                                                        ace_commands.append(
                                                            "no "
                                                            + str(
                                                                ace["sequence"]
                                                            )
                                                        )
                                                        flag = 1
                                        else:
                                            if "aces" in have_name.keys():
                                                for h_ace in have_name["aces"]:
                                                    # when want['ace'] does not have seq number
                                                    if (
                                                        "sequence"
                                                        not in ace.keys()
                                                    ):
                                                        del h_ace["sequence"]
                                                    if ace == h_ace:
                                                        ace_commands.append(
                                                            "no "
                                                            + self.process_ace(
                                                                ace
                                                            )
                                                        )
                                                        flag = 1
                                    if flag:
                                        ace_commands.insert(
                                            0,
                                            ip + " access-list " + acl["name"],
                                        )
                                    commands.extend(ace_commands)
                            else:
                                # only name given
                                for h in have_afi["acls"]:
                                    if h["name"] == acl["name"]:
                                        acl_names.append(acl["name"])
                        for name in acl_names:
                            commands.append(
                                "no " + ip + " access-list " + name
                            )

                    else:
                        # 'only afi is given'
                        if have_afi.get("acls"):
                            for h in have_afi["acls"]:
                                acl_names.append(h["name"])
                            for name in acl_names:
                                commands.append(
                                    "no " + ip + " access-list " + name
                                )
        else:
            v6 = []
            v4 = []
            v6_local = v4_local = None
            for h in have:
                if h["afi"] == "ipv6":
                    v6 = (acl["name"] for acl in h["acls"])
                    if "match_local_traffic" in h.keys():
                        v6_local = True
                else:
                    v4 = (acl["name"] for acl in h["acls"])
                    if "match_local_traffic" in h.keys():
                        v4_local = True

            self.no_commands(v4, commands, v4_local, "ip")
            self.no_commands(v6, commands, v6_local, "ipv6")

            for name in v6:
                commands.append("no ipv6 access-list " + name)
            if v4_local:
                commands.append("no ipv6 access-list match-local-traffic")

        return commands

    def no_commands(self, v_list, commands, match_local, ip):
        for name in v_list:
            commands.append("no " + ip + " access-list " + name)
        if match_local:
            commands.append("no " + ip + " access-list match-local-traffic")

    def set_commands(self, want, have):
        commands = []
        have_afi = search_obj_in_list(want["afi"], have, "afi")
        ip = ""
        if "v6" in want["afi"]:
            ip = "ipv6 "
        else:
            ip = "ip "

        if have_afi:
            if want.get("acls"):
                for w_acl in want["acls"]:
                    have_acl = search_obj_in_list(
                        w_acl["name"], have_afi["acls"], "name"
                    )
                    name = w_acl["name"]
                    flag = 0
                    ace_commands = []
                    if have_acl != w_acl:
                        if have_acl:
                            ace_list = []
                            if w_acl.get("aces") and have_acl.get("aces"):
                                # case 1 --> sequence number not given in want --> new ace
                                # case 2 --> new sequence number in want --> new ace
                                # case 3 --> existing sequence number given --> update rule (only for merged state.
                                #            For replaced and overridden, rule is deleted in the state's config)

                                ace_list = [
                                    item
                                    for item in w_acl["aces"]
                                    if "sequence" not in item.keys()
                                ]  # case 1

                                want_seq = [
                                    item["sequence"]
                                    for item in w_acl["aces"]
                                    if "sequence" in item.keys()
                                ]

                                have_seq = [
                                    item["sequence"]
                                    for item in have_acl["aces"]
                                ]

                                new_seq = list(set(want_seq) - set(have_seq))
                                common_seq = list(
                                    set(want_seq).intersection(set(have_seq))
                                )

                                temp_list = [
                                    item
                                    for item in w_acl["aces"]
                                    if "sequence" in item.keys()
                                    and item["sequence"] in new_seq
                                ]  # case 2
                                ace_list.extend(temp_list)
                                for w in w_acl["aces"]:
                                    self.argument_spec = AclsArgs.argument_spec
                                    params = utils.validate_config(
                                        self.argument_spec,
                                        {
                                            "config": [
                                                {
                                                    "afi": want["afi"],
                                                    "acls": [
                                                        {
                                                            "name": name,
                                                            "aces": ace_list,
                                                        }
                                                    ],
                                                }
                                            ]
                                        },
                                    )
                                    if (
                                        "sequence" in w.keys()
                                        and w["sequence"] in common_seq
                                    ):
                                        temp_obj = search_obj_in_list(
                                            w["sequence"],
                                            have_acl["aces"],
                                            "sequence",
                                        )  # case 3
                                        if temp_obj != w:
                                            for key, val in w.items():
                                                temp_obj[key] = val
                                            ace_list.append(temp_obj)
                                            if (
                                                self._module.params["state"]
                                                == "merged"
                                            ):
                                                ace_commands.append(
                                                    "no " + str(w["sequence"])
                                                )
                                        # remove existing rule to update it
                            elif w_acl.get("aces"):
                                # 'have' has ACL defined without any ACE
                                ace_list = [item for item in w_acl["aces"]]
                            for w_ace in ace_list:
                                ace_commands.append(
                                    self.process_ace(w_ace).strip()
                                )
                                flag = 1

                            if flag:
                                ace_commands.insert(
                                    0, ip + "access-list " + name
                                )

                        else:
                            commands.append(ip + "access-list " + name)
                            if "aces" in w_acl.keys():
                                for w_ace in w_acl["aces"]:
                                    commands.append(
                                        self.process_ace(w_ace).strip()
                                    )
                    commands.extend(ace_commands)
        else:
            if want.get("acls"):
                for w_acl in want["acls"]:
                    name = w_acl["name"]
                    commands.append(ip + "access-list " + name)
                    if "aces" in w_acl.keys():
                        for w_ace in w_acl["aces"]:
                            commands.append(self.process_ace(w_ace).strip())

        return commands

    def process_ace(self, w_ace):
        command = ""
        ace_keys = w_ace.keys()
        if "remark" in ace_keys:
            command += "remark " + w_ace["remark"] + " "
        else:
            command += w_ace["grant"] + " "
            if "protocol" in ace_keys:
                command += w_ace["protocol"] + " "
                src = self.get_address(w_ace["source"], w_ace["protocol"])
                dest = self.get_address(
                    w_ace["destination"], w_ace["protocol"]
                )
                command += src + dest
                if "protocol_options" in ace_keys:
                    pro = list(w_ace["protocol_options"].keys())[0]
                    if pro != w_ace["protocol"]:
                        self._module.fail_json(
                            msg="protocol and protocol_options mismatch"
                        )
                    flags = ""
                    for k in w_ace["protocol_options"][pro].keys():
                        k = re.sub("_", "-", k)
                        flags += k + " "
                    command += flags
                if "dscp" in ace_keys:
                    command += "dscp " + w_ace["dscp"] + " "
                if "fragments" in ace_keys:
                    command += "fragments "
                if "precedence" in ace_keys:
                    command += "precedence " + w_ace["precedence"] + " "
            if "log" in ace_keys:
                command += "log "
        if "sequence" in ace_keys:
            command = str(w_ace["sequence"]) + " " + command
        return command

    def get_address(self, endpoint, pro=""):
        ret_addr = ""
        keys = list(endpoint.keys())
        if "address" in keys:
            if "wildcard_bits" not in keys:
                self._module.fail_json(
                    msg="wildcard bits not specified for address"
                )
            else:
                ret_addr = (
                    endpoint["address"] + " " + endpoint["wildcard_bits"] + " "
                )
        elif "any" in keys:
            ret_addr = "any "
        elif "host" in keys:
            ret_addr = "host " + endpoint["host"] + " "
        elif "prefix" in keys:
            ret_addr = endpoint["prefix"] + " "

        if pro in ["tcp", "udp"]:
            if "port_protocol" in keys:
                options = self.get_options(endpoint["port_protocol"])
                ret_addr += options
        return ret_addr

    def get_options(self, item):
        com = ""
        subkey = list(item.keys())
        if "range" in subkey:
            com = (
                "range "
                + item["range"]["start"]
                + " "
                + item["range"]["end"]
                + " "
            )
        else:
            com = subkey[0] + " " + item[subkey[0]] + " "
        return com



# This excerpt is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file at
# https://github.com/pyca/cryptography/blob/master/LICENSE for complete details.
#
# Adapted from cryptography's hazmat/backends/openssl/decode_asn1.py
#
# Copyright (c) 2015, 2016 Paul Kehrer (@reaperhulk)
# Copyright (c) 2017 Fraser Tweedale (@frasertweedale)

# Relevant commits from cryptography project (https://github.com/pyca/cryptography):
#    pyca/cryptography@719d536dd691e84e208534798f2eb4f82aaa2e07
#    pyca/cryptography@5ab6d6a5c05572bd1c75f05baf264a2d0001894a
#    pyca/cryptography@2e776e20eb60378e0af9b7439000d0e80da7c7e3
#    pyca/cryptography@fb309ed24647d1be9e319b61b1f2aa8ebb87b90b
#    pyca/cryptography@2917e460993c475c72d7146c50dc3bbc2414280d
#    pyca/cryptography@3057f91ea9a05fb593825006d87a391286a4d828
#    pyca/cryptography@d607dd7e5bc5c08854ec0c9baff70ba4a35be36f

from __future__ import absolute_import, division, print_function
__metaclass__ = type


def obj2txt(openssl_lib, openssl_ffi, obj):
    # Set to 80 on the recommendation of
    # https://www.openssl.org/docs/crypto/OBJ_nid2ln.html#return_values
    #
    # But OIDs longer than this occur in real life (e.g. Active
    # Directory makes some very long OIDs).  So we need to detect
    # and properly handle the case where the default buffer is not
    # big enough.
    #
    buf_len = 80
    buf = openssl_ffi.new("char[]", buf_len)

    # 'res' is the number of bytes that *would* be written if the
    # buffer is large enough.  If 'res' > buf_len - 1, we need to
    # alloc a big-enough buffer and go again.
    res = openssl_lib.OBJ_obj2txt(buf, buf_len, obj, 1)
    if res > buf_len - 1:  # account for terminating null byte
        buf_len = res + 1
        buf = openssl_ffi.new("char[]", buf_len)
        res = openssl_lib.OBJ_obj2txt(buf, buf_len, obj, 1)
    return openssl_ffi.buffer(buf, res)[:].decode()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: vertica_info
short_description: Gathers Vertica database facts.
description:
  - Gathers Vertica database information.
  - This module was called C(vertica_facts) before Ansible 2.9, returning C(ansible_facts).
    Note that the M(community.general.vertica_info) module no longer returns C(ansible_facts)!
options:
  cluster:
    description:
      - Name of the cluster running the schema.
    default: localhost
  port:
    description:
      Database port to connect to.
    default: 5433
  db:
    description:
      - Name of the database running the schema.
  login_user:
    description:
      - The username used to authenticate with.
    default: dbadmin
  login_password:
    description:
      - The password used to authenticate with.
notes:
  - The default authentication assumes that you are either logging in as or sudo'ing
    to the C(dbadmin) account on the host.
  - This module uses C(pyodbc), a Python ODBC database adapter. You must ensure
    that C(unixODBC) and C(pyodbc) is installed on the host and properly configured.
  - Configuring C(unixODBC) for Vertica requires C(Driver = /opt/vertica/lib64/libverticaodbc.so)
    to be added to the C(Vertica) section of either C(/etc/odbcinst.ini) or C($HOME/.odbcinst.ini)
    and both C(ErrorMessagesPath = /opt/vertica/lib64) and C(DriverManagerEncoding = UTF-16)
    to be added to the C(Driver) section of either C(/etc/vertica.ini) or C($HOME/.vertica.ini).
requirements: [ 'unixODBC', 'pyodbc' ]
author: "Dariusz Owczarek (@dareko)"
'''

EXAMPLES = """
- name: Gathering vertica facts
  community.general.vertica_info: db=db_name
  register: result

- name: Print schemas
  ansible.builtin.debug:
    msg: "{{ result.vertica_schemas }}"
"""
import traceback

PYODBC_IMP_ERR = None
try:
    import pyodbc
except ImportError:
    PYODBC_IMP_ERR = traceback.format_exc()
    pyodbc_found = False
else:
    pyodbc_found = True

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils._text import to_native


class NotSupportedError(Exception):
    pass

# module specific functions


def get_schema_facts(cursor, schema=''):
    facts = {}
    cursor.execute("""
        select schema_name, schema_owner, create_time
        from schemata
        where not is_system_schema and schema_name not in ('public')
        and (? = '' or schema_name ilike ?)
    """, schema, schema)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            facts[row.schema_name.lower()] = {
                'name': row.schema_name,
                'owner': row.schema_owner,
                'create_time': str(row.create_time),
                'usage_roles': [],
                'create_roles': []}
    cursor.execute("""
        select g.object_name as schema_name, r.name as role_name,
        lower(g.privileges_description) privileges_description
        from roles r join grants g
        on g.grantee = r.name and g.object_type='SCHEMA'
        and g.privileges_description like '%USAGE%'
        and g.grantee not in ('public', 'dbadmin')
        and (? = '' or g.object_name ilike ?)
    """, schema, schema)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            schema_key = row.schema_name.lower()
            if 'create' in row.privileges_description:
                facts[schema_key]['create_roles'].append(row.role_name)
            else:
                facts[schema_key]['usage_roles'].append(row.role_name)
    return facts


def get_user_facts(cursor, user=''):
    facts = {}
    cursor.execute("""
        select u.user_name, u.is_locked, u.lock_time,
        p.password, p.acctexpired as is_expired,
        u.profile_name, u.resource_pool,
        u.all_roles, u.default_roles
        from users u join password_auditor p on p.user_id = u.user_id
        where not u.is_super_user
        and (? = '' or u.user_name ilike ?)
     """, user, user)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            user_key = row.user_name.lower()
            facts[user_key] = {
                'name': row.user_name,
                'locked': str(row.is_locked),
                'password': row.password,
                'expired': str(row.is_expired),
                'profile': row.profile_name,
                'resource_pool': row.resource_pool,
                'roles': [],
                'default_roles': []}
            if row.is_locked:
                facts[user_key]['locked_time'] = str(row.lock_time)
            if row.all_roles:
                facts[user_key]['roles'] = row.all_roles.replace(' ', '').split(',')
            if row.default_roles:
                facts[user_key]['default_roles'] = row.default_roles.replace(' ', '').split(',')
    return facts


def get_role_facts(cursor, role=''):
    facts = {}
    cursor.execute("""
        select r.name, r.assigned_roles
        from roles r
        where (? = '' or r.name ilike ?)
    """, role, role)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            role_key = row.name.lower()
            facts[role_key] = {
                'name': row.name,
                'assigned_roles': []}
            if row.assigned_roles:
                facts[role_key]['assigned_roles'] = row.assigned_roles.replace(' ', '').split(',')
    return facts


def get_configuration_facts(cursor, parameter=''):
    facts = {}
    cursor.execute("""
        select c.parameter_name, c.current_value, c.default_value
        from configuration_parameters c
        where c.node_name = 'ALL'
        and (? = '' or c.parameter_name ilike ?)
    """, parameter, parameter)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            facts[row.parameter_name.lower()] = {
                'parameter_name': row.parameter_name,
                'current_value': row.current_value,
                'default_value': row.default_value}
    return facts


def get_node_facts(cursor, schema=''):
    facts = {}
    cursor.execute("""
        select node_name, node_address, export_address, node_state, node_type,
            catalog_path
        from nodes
    """)
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            facts[row.node_address] = {
                'node_name': row.node_name,
                'export_address': row.export_address,
                'node_state': row.node_state,
                'node_type': row.node_type,
                'catalog_path': row.catalog_path}
    return facts

# module logic


def main():

    module = AnsibleModule(
        argument_spec=dict(
            cluster=dict(default='localhost'),
            port=dict(default='5433'),
            db=dict(default=None),
            login_user=dict(default='dbadmin'),
            login_password=dict(default=None, no_log=True),
        ), supports_check_mode=True)
    is_old_facts = module._name in ('vertica_facts', 'community.general.vertica_facts')
    if is_old_facts:
        module.deprecate("The 'vertica_facts' module has been renamed to 'vertica_info', "
                         "and the renamed one no longer returns ansible_facts",
                         version='3.0.0', collection_name='community.general')  # was Ansible 2.13

    if not pyodbc_found:
        module.fail_json(msg=missing_required_lib('pyodbc'), exception=PYODBC_IMP_ERR)

    db = ''
    if module.params['db']:
        db = module.params['db']

    try:
        dsn = (
            "Driver=Vertica;"
            "Server=%s;"
            "Port=%s;"
            "Database=%s;"
            "User=%s;"
            "Password=%s;"
            "ConnectionLoadBalance=%s"
        ) % (module.params['cluster'], module.params['port'], db,
             module.params['login_user'], module.params['login_password'], 'true')
        db_conn = pyodbc.connect(dsn, autocommit=True)
        cursor = db_conn.cursor()
    except Exception as e:
        module.fail_json(msg="Unable to connect to database: %s." % to_native(e), exception=traceback.format_exc())

    try:
        schema_facts = get_schema_facts(cursor)
        user_facts = get_user_facts(cursor)
        role_facts = get_role_facts(cursor)
        configuration_facts = get_configuration_facts(cursor)
        node_facts = get_node_facts(cursor)

        if is_old_facts:
            module.exit_json(changed=False,
                             ansible_facts={'vertica_schemas': schema_facts,
                                            'vertica_users': user_facts,
                                            'vertica_roles': role_facts,
                                            'vertica_configuration': configuration_facts,
                                            'vertica_nodes': node_facts})
        else:
            module.exit_json(changed=False,
                             vertica_schemas=schema_facts,
                             vertica_users=user_facts,
                             vertica_roles=role_facts,
                             vertica_configuration=configuration_facts,
                             vertica_nodes=node_facts)
    except NotSupportedError as e:
        module.fail_json(msg=to_native(e), exception=traceback.format_exc())
    except SystemExit:
        # avoid catching this on python 2.4
        raise
    except Exception as e:
        module.fail_json(msg=to_native(e), exception=traceback.format_exc())


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2014, Kevin Carter <kevin.carter@rackspace.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: lxc_container
short_description: Manage LXC Containers
description:
  - Management of LXC containers.
author: "Kevin Carter (@cloudnull)"
options:
    name:
        description:
          - Name of a container.
        type: str
        required: true
    backing_store:
        choices:
          - dir
          - lvm
          - loop
          - btrfs
          - overlayfs
          - zfs
        description:
          - Backend storage type for the container.
        type: str
        default: dir
    template:
        description:
          - Name of the template to use within an LXC create.
        type: str
        default: ubuntu
    template_options:
        description:
          - Template options when building the container.
        type: str
    config:
        description:
          - Path to the LXC configuration file.
        type: path
    lv_name:
        description:
          - Name of the logical volume, defaults to the container name.
          - If not specified, it defaults to C($CONTAINER_NAME).
        type: str
    vg_name:
        description:
          - If backend store is lvm, specify the name of the volume group.
        type: str
        default: lxc
    thinpool:
        description:
          - Use LVM thin pool called TP.
        type: str
    fs_type:
        description:
          - Create fstype TYPE.
        type: str
        default: ext4
    fs_size:
        description:
          - File system Size.
        type: str
        default: 5G
    directory:
        description:
          - Place rootfs directory under DIR.
        type: path
    zfs_root:
        description:
          - Create zfs under given zfsroot.
        type: str
    container_command:
        description:
          - Run a command within a container.
        type: str
    lxc_path:
        description:
          - Place container under PATH.
        type: path
    container_log:
        description:
          - Enable a container log for host actions to the container.
        type: bool
        default: 'no'
    container_log_level:
        choices:
          - Info
          - info
          - INFO
          - Error
          - error
          - ERROR
          - Debug
          - debug
          - DEBUG
        description:
          - Set the log level for a container where *container_log* was set.
        type: str
        required: false
        default: INFO
    clone_name:
        description:
          - Name of the new cloned server.
          - This is only used when state is clone.
        type: str
    clone_snapshot:
        description:
          - Create a snapshot a container when cloning.
          - This is not supported by all container storage backends.
          - Enabling this may fail if the backing store does not support snapshots.
        type: bool
        default: 'no'
    archive:
        description:
          - Create an archive of a container.
          - This will create a tarball of the running container.
        type: bool
        default: 'no'
    archive_path:
        description:
          - Path the save the archived container.
          - If the path does not exist the archive method will attempt to create it.
        type: path
    archive_compression:
        choices:
          - gzip
          - bzip2
          - none
        description:
          - Type of compression to use when creating an archive of a running
            container.
        type: str
        default: gzip
    state:
        choices:
          - started
          - stopped
          - restarted
          - absent
          - frozen
          - clone
        description:
          - Define the state of a container.
          - If you clone a container using I(clone_name) the newly cloned
            container created in a stopped state.
          - The running container will be stopped while the clone operation is
            happening and upon completion of the clone the original container
            state will be restored.
        type: str
        default: started
    container_config:
        description:
          - A list of C(key=value) options to use when configuring a container.
        type: list
        elements: str
requirements:
  - 'lxc >= 1.0 # OS package'
  - 'python >= 2.6 # OS Package'
  - 'lxc-python2 >= 0.1 # PIP Package from https://github.com/lxc/python2-lxc'
notes:
  - Containers must have a unique name. If you attempt to create a container
    with a name that already exists in the users namespace the module will
    simply return as "unchanged".
  - The "container_command" can be used with any state except "absent". If
    used with state "stopped" the container will be "started", the command
    executed, and then the container "stopped" again. Likewise if the state
    is "stopped" and the container does not exist it will be first created,
    "started", the command executed, and then "stopped". If you use a "|"
    in the variable you can use common script formatting within the variable
    itself The "container_command" option will always execute as BASH.
    When using "container_command" a log file is created in the /tmp/ directory
    which contains both stdout and stderr of any command executed.
  - If "archive" is **true** the system will attempt to create a compressed
    tarball of the running container. The "archive" option supports LVM backed
    containers and will create a snapshot of the running container when
    creating the archive.
  - If your distro does not have a package for "python2-lxc", which is a
    requirement for this module, it can be installed from source at
    "https://github.com/lxc/python2-lxc" or installed via pip using the package
    name lxc-python2.
'''

EXAMPLES = r"""
- name: Create a started container
  community.general.lxc_container:
    name: test-container-started
    container_log: true
    template: ubuntu
    state: started
    template_options: --release trusty

- name: Create a stopped container
  community.general.lxc_container:
    name: test-container-stopped
    container_log: true
    template: ubuntu
    state: stopped
    template_options: --release trusty

- name: Create a frozen container
  community.general.lxc_container:
    name: test-container-frozen
    container_log: true
    template: ubuntu
    state: frozen
    template_options: --release trusty
    container_command: |
      echo 'hello world.' | tee /opt/started-frozen

# Create filesystem container, configure it, and archive it, and start it.
- name: Create filesystem container
  community.general.lxc_container:
    name: test-container-config
    backing_store: dir
    container_log: true
    template: ubuntu
    state: started
    archive: true
    archive_compression: none
    container_config:
      - "lxc.aa_profile=unconfined"
      - "lxc.cgroup.devices.allow=a *:* rmw"
    template_options: --release trusty

# Create an lvm container, run a complex command in it, add additional
# configuration to it, create an archive of it, and finally leave the container
# in a frozen state. The container archive will be compressed using bzip2
- name: Create a frozen lvm container
  community.general.lxc_container:
    name: test-container-lvm
    container_log: true
    template: ubuntu
    state: frozen
    backing_store: lvm
    template_options: --release trusty
    container_command: |
      apt-get update
      apt-get install -y vim lxc-dev
      echo 'hello world.' | tee /opt/started
      if [[ -f "/opt/started" ]]; then
          echo 'hello world.' | tee /opt/found-started
      fi
    container_config:
      - "lxc.aa_profile=unconfined"
      - "lxc.cgroup.devices.allow=a *:* rmw"
    archive: true
    archive_compression: bzip2
  register: lvm_container_info

- name: Debug info on container "test-container-lvm"
  ansible.builtin.debug:
    var: lvm_container_info

- name: Run a command in a container and ensure its in a "stopped" state.
  community.general.lxc_container:
    name: test-container-started
    state: stopped
    container_command: |
      echo 'hello world.' | tee /opt/stopped

- name: Run a command in a container and ensure its it in a "frozen" state.
  community.general.lxc_container:
    name: test-container-stopped
    state: frozen
    container_command: |
      echo 'hello world.' | tee /opt/frozen

- name: Start a container
  community.general.lxc_container:
    name: test-container-stopped
    state: started

- name: Run a command in a container and then restart it
  community.general.lxc_container:
    name: test-container-started
    state: restarted
    container_command: |
      echo 'hello world.' | tee /opt/restarted

- name: Run a complex command within a "running" container
  community.general.lxc_container:
    name: test-container-started
    container_command: |
      apt-get update
      apt-get install -y curl wget vim apache2
      echo 'hello world.' | tee /opt/started
      if [[ -f "/opt/started" ]]; then
          echo 'hello world.' | tee /opt/found-started
      fi

# Create an archive of an existing container, save the archive to a defined
# path and then destroy it.
- name: Archive container
  community.general.lxc_container:
    name: test-container-started
    state: absent
    archive: true
    archive_path: /opt/archives

# Create a container using overlayfs, create an archive of it, create a
# snapshot clone of the container and and finally leave the container
# in a frozen state. The container archive will be compressed using gzip.
- name: Create an overlayfs container archive and clone it
  community.general.lxc_container:
    name: test-container-overlayfs
    container_log: true
    template: ubuntu
    state: started
    backing_store: overlayfs
    template_options: --release trusty
    clone_snapshot: true
    clone_name: test-container-overlayfs-clone-snapshot
    archive: true
    archive_compression: gzip
  register: clone_container_info

- name: Debug info on container "test-container"
  ansible.builtin.debug:
    var: clone_container_info

- name: Clone a container using snapshot
  community.general.lxc_container:
    name: test-container-overlayfs-clone-snapshot
    backing_store: overlayfs
    clone_name: test-container-overlayfs-clone-snapshot2
    clone_snapshot: true

- name: Create a new container and clone it
  community.general.lxc_container:
    name: test-container-new-archive
    backing_store: dir
    clone_name: test-container-new-archive-clone

- name: Archive and clone a container then destroy it
  community.general.lxc_container:
    name: test-container-new-archive
    state: absent
    clone_name: test-container-new-archive-destroyed-clone
    archive: true
    archive_compression: gzip

- name: Start a cloned container.
  community.general.lxc_container:
    name: test-container-new-archive-destroyed-clone
    state: started

- name: Destroy a container
  community.general.lxc_container:
    name: '{{ item }}'
    state: absent
  with_items:
    - test-container-stopped
    - test-container-started
    - test-container-frozen
    - test-container-lvm
    - test-container-config
    - test-container-overlayfs
    - test-container-overlayfs-clone
    - test-container-overlayfs-clone-snapshot
    - test-container-overlayfs-clone-snapshot2
    - test-container-new-archive
    - test-container-new-archive-clone
    - test-container-new-archive-destroyed-clone
"""

RETURN = r"""
lxc_container:
    description: container information
    returned: success
    type: complex
    contains:
        name:
            description: name of the lxc container
            returned: success
            type: str
            sample: test_host
        init_pid:
            description: pid of the lxc init process
            returned: success
            type: int
            sample: 19786
        interfaces:
            description: list of the container's network interfaces
            returned: success
            type: list
            sample: [ "eth0", "lo" ]
        ips:
            description: list of ips
            returned: success
            type: list
            sample: [ "10.0.3.3" ]
        state:
            description: resulting state of the container
            returned: success
            type: str
            sample: "running"
        archive:
            description: resulting state of the container
            returned: success, when archive is true
            type: str
            sample: "/tmp/test-container-config.tar"
        clone:
            description: if the container was cloned
            returned: success, when clone_name is specified
            type: bool
            sample: True
"""

import os
import os.path
import re
import shutil
import subprocess
import tempfile
import time

try:
    import lxc
except ImportError:
    HAS_LXC = False
else:
    HAS_LXC = True

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.parsing.convert_bool import BOOLEANS_FALSE, BOOLEANS_TRUE
from ansible.module_utils.six.moves import xrange
from ansible.module_utils._text import to_text, to_bytes


# LXC_COMPRESSION_MAP is a map of available compression types when creating
# an archive of a container.
LXC_COMPRESSION_MAP = {
    'gzip': {
        'extension': 'tar.tgz',
        'argument': '-czf'
    },
    'bzip2': {
        'extension': 'tar.bz2',
        'argument': '-cjf'
    },
    'none': {
        'extension': 'tar',
        'argument': '-cf'
    }
}


# LXC_COMMAND_MAP is a map of variables that are available to a method based
# on the state the container is in.
LXC_COMMAND_MAP = {
    'create': {
        'variables': {
            'config': '--config',
            'template': '--template',
            'backing_store': '--bdev',
            'lxc_path': '--lxcpath',
            'lv_name': '--lvname',
            'vg_name': '--vgname',
            'thinpool': '--thinpool',
            'fs_type': '--fstype',
            'fs_size': '--fssize',
            'directory': '--dir',
            'zfs_root': '--zfsroot'
        }
    },
    'clone': {
        'variables-lxc-copy': {
            'backing_store': '--backingstorage',
            'lxc_path': '--lxcpath',
            'fs_size': '--fssize',
            'name': '--name',
            'clone_name': '--newname'
        },
        # lxc-clone is deprecated in favor of lxc-copy
        'variables-lxc-clone': {
            'backing_store': '--backingstore',
            'lxc_path': '--lxcpath',
            'fs_size': '--fssize',
            'name': '--orig',
            'clone_name': '--new'
        }
    }
}


# LXC_BACKING_STORE is a map of available storage backends and options that
# are incompatible with the given storage backend.
LXC_BACKING_STORE = {
    'dir': [
        'lv_name', 'vg_name', 'fs_type', 'fs_size', 'thinpool'
    ],
    'lvm': [
        'zfs_root'
    ],
    'btrfs': [
        'lv_name', 'vg_name', 'thinpool', 'zfs_root', 'fs_type', 'fs_size'
    ],
    'loop': [
        'lv_name', 'vg_name', 'thinpool', 'zfs_root'
    ],
    'overlayfs': [
        'lv_name', 'vg_name', 'fs_type', 'fs_size', 'thinpool', 'zfs_root'
    ],
    'zfs': [
        'lv_name', 'vg_name', 'fs_type', 'fs_size', 'thinpool'
    ]
}


# LXC_LOGGING_LEVELS is a map of available log levels
LXC_LOGGING_LEVELS = {
    'INFO': ['info', 'INFO', 'Info'],
    'ERROR': ['error', 'ERROR', 'Error'],
    'DEBUG': ['debug', 'DEBUG', 'Debug']
}


# LXC_ANSIBLE_STATES is a map of states that contain values of methods used
# when a particular state is evoked.
LXC_ANSIBLE_STATES = {
    'started': '_started',
    'stopped': '_stopped',
    'restarted': '_restarted',
    'absent': '_destroyed',
    'frozen': '_frozen',
    'clone': '_clone'
}


# This is used to attach to a running container and execute commands from
# within the container on the host.  This will provide local access to a
# container without using SSH.  The template will attempt to work within the
# home directory of the user that was attached to the container and source
# that users environment variables by default.
ATTACH_TEMPLATE = """#!/usr/bin/env bash
pushd "$(getent passwd $(whoami)|cut -f6 -d':')"
    if [[ -f ".bashrc" ]];then
        source .bashrc
        unset HOSTNAME
    fi
popd

# User defined command
%(container_command)s
"""


def create_script(command):
    """Write out a script onto a target.

    This method should be backward compatible with Python 2.4+ when executing
    from within the container.

    :param command: command to run, this can be a script and can use spacing
                    with newlines as separation.
    :type command: ``str``
    """

    (fd, script_file) = tempfile.mkstemp(prefix='lxc-attach-script')
    f = os.fdopen(fd, 'wb')
    try:
        f.write(to_bytes(ATTACH_TEMPLATE % {'container_command': command}, errors='surrogate_or_strict'))
        f.flush()
    finally:
        f.close()

    # Ensure the script is executable.
    os.chmod(script_file, int('0700', 8))

    # Output log file.
    stdout_file = os.fdopen(tempfile.mkstemp(prefix='lxc-attach-script-log')[0], 'ab')

    # Error log file.
    stderr_file = os.fdopen(tempfile.mkstemp(prefix='lxc-attach-script-err')[0], 'ab')

    # Execute the script command.
    try:
        subprocess.Popen(
            [script_file],
            stdout=stdout_file,
            stderr=stderr_file
        ).communicate()
    finally:
        # Close the log files.
        stderr_file.close()
        stdout_file.close()

        # Remove the script file upon completion of execution.
        os.remove(script_file)


class LxcContainerManagement(object):
    def __init__(self, module):
        """Management of LXC containers via Ansible.

        :param module: Processed Ansible Module.
        :type module: ``object``
        """
        self.module = module
        self.state = self.module.params.get('state', None)
        self.state_change = False
        self.lxc_vg = None
        self.lxc_path = self.module.params.get('lxc_path', None)
        self.container_name = self.module.params['name']
        self.container = self.get_container_bind()
        self.archive_info = None
        self.clone_info = None

    def get_container_bind(self):
        return lxc.Container(name=self.container_name)

    @staticmethod
    def _roundup(num):
        """Return a rounded floating point number.

        :param num: Number to round up.
        :type: ``float``
        :returns: Rounded up number.
        :rtype: ``int``
        """
        num, part = str(num).split('.')
        num = int(num)
        if int(part) != 0:
            num += 1
        return num

    @staticmethod
    def _container_exists(container_name, lxc_path=None):
        """Check if a container exists.

        :param container_name: Name of the container.
        :type: ``str``
        :returns: True or False if the container is found.
        :rtype: ``bol``
        """
        if [i for i in lxc.list_containers(config_path=lxc_path) if i == container_name]:
            return True
        else:
            return False

    @staticmethod
    def _add_variables(variables_dict, build_command):
        """Return a command list with all found options.

        :param variables_dict: Pre-parsed optional variables used from a
                               seed command.
        :type variables_dict: ``dict``
        :param build_command: Command to run.
        :type build_command: ``list``
        :returns: list of command options.
        :rtype: ``list``
        """

        for key, value in variables_dict.items():
            build_command.append(
                '%s %s' % (key, value)
            )
        return build_command

    def _get_vars(self, variables):
        """Return a dict of all variables as found within the module.

        :param variables: Hash of all variables to find.
        :type variables: ``dict``
        """

        # Remove incompatible storage backend options.
        variables = variables.copy()
        for v in LXC_BACKING_STORE[self.module.params['backing_store']]:
            variables.pop(v, None)

        return_dict = dict()
        false_values = BOOLEANS_FALSE.union([None, ''])
        for k, v in variables.items():
            _var = self.module.params.get(k)
            if _var not in false_values:
                return_dict[v] = _var
        return return_dict

    def _run_command(self, build_command, unsafe_shell=False):
        """Return information from running an Ansible Command.

        This will squash the build command list into a string and then
        execute the command via Ansible. The output is returned to the method.
        This output is returned as `return_code`, `stdout`, `stderr`.

        :param build_command: Used for the command and all options.
        :type build_command: ``list``
        :param unsafe_shell: Enable or Disable unsafe sell commands.
        :type unsafe_shell: ``bol``
        """

        return self.module.run_command(
            ' '.join(build_command),
            use_unsafe_shell=unsafe_shell
        )

    def _config(self):
        """Configure an LXC container.

        Write new configuration values to the lxc config file. This will
        stop the container if it's running write the new options and then
        restart the container upon completion.
        """

        _container_config = self.module.params.get('container_config')
        if not _container_config:
            return False

        container_config_file = self.container.config_file_name
        with open(container_config_file, 'rb') as f:
            container_config = to_text(f.read(), errors='surrogate_or_strict').splitlines(True)

        parsed_options = [i.split('=', 1) for i in _container_config]
        config_change = False
        for key, value in parsed_options:
            key = key.strip()
            value = value.strip()
            new_entry = '%s = %s\n' % (key, value)
            keyre = re.compile(r'%s(\s+)?=' % key)
            for option_line in container_config:
                # Look for key in config
                if keyre.match(option_line):
                    _, _value = option_line.split('=', 1)
                    config_value = ' '.join(_value.split())
                    line_index = container_config.index(option_line)
                    # If the sanitized values don't match replace them
                    if value != config_value:
                        line_index += 1
                        if new_entry not in container_config:
                            config_change = True
                            container_config.insert(line_index, new_entry)
                    # Break the flow as values are written or not at this point
                    break
            else:
                config_change = True
                container_config.append(new_entry)

        # If the config changed restart the container.
        if config_change:
            container_state = self._get_state()
            if container_state != 'stopped':
                self.container.stop()

            with open(container_config_file, 'wb') as f:
                f.writelines([to_bytes(line, errors='surrogate_or_strict') for line in container_config])

            self.state_change = True
            if container_state == 'running':
                self._container_startup()
            elif container_state == 'frozen':
                self._container_startup()
                self.container.freeze()

    def _container_create_clone(self):
        """Clone a new LXC container from an existing container.

        This method will clone an existing container to a new container using
        the `clone_name` variable as the new container name. The method will
        create a container if the container `name` does not exist.

        Note that cloning a container will ensure that the original container
        is "stopped" before the clone can be done. Because this operation can
        require a state change the method will return the original container
        to its prior state upon completion of the clone.

        Once the clone is complete the new container will be left in a stopped
        state.
        """

        # Ensure that the state of the original container is stopped
        container_state = self._get_state()
        if container_state != 'stopped':
            self.state_change = True
            self.container.stop()

        # lxc-clone is deprecated in favor of lxc-copy
        clone_vars = 'variables-lxc-copy'
        clone_cmd = self.module.get_bin_path('lxc-copy')
        if not clone_cmd:
            clone_vars = 'variables-lxc-clone'
            clone_cmd = self.module.get_bin_path('lxc-clone', True)

        build_command = [
            clone_cmd,
        ]

        build_command = self._add_variables(
            variables_dict=self._get_vars(
                variables=LXC_COMMAND_MAP['clone'][clone_vars]
            ),
            build_command=build_command
        )

        # Load logging for the instance when creating it.
        if self.module.params.get('clone_snapshot') in BOOLEANS_TRUE:
            build_command.append('--snapshot')
        # Check for backing_store == overlayfs if so force the use of snapshot
        # If overlay fs is used and snapshot is unset the clone command will
        # fail with an unsupported type.
        elif self.module.params.get('backing_store') == 'overlayfs':
            build_command.append('--snapshot')

        rc, return_data, err = self._run_command(build_command)
        if rc != 0:
            message = "Failed executing %s." % os.path.basename(clone_cmd)
            self.failure(
                err=err, rc=rc, msg=message, command=' '.join(
                    build_command
                )
            )
        else:
            self.state_change = True
            # Restore the original state of the origin container if it was
            # not in a stopped state.
            if container_state == 'running':
                self.container.start()
            elif container_state == 'frozen':
                self.container.start()
                self.container.freeze()

        return True

    def _create(self):
        """Create a new LXC container.

        This method will build and execute a shell command to build the
        container. It would have been nice to simply use the lxc python library
        however at the time this was written the python library, in both py2
        and py3 didn't support some of the more advanced container create
        processes. These missing processes mainly revolve around backing
        LXC containers with block devices.
        """

        build_command = [
            self.module.get_bin_path('lxc-create', True),
            '--name %s' % self.container_name,
            '--quiet'
        ]

        build_command = self._add_variables(
            variables_dict=self._get_vars(
                variables=LXC_COMMAND_MAP['create']['variables']
            ),
            build_command=build_command
        )

        # Load logging for the instance when creating it.
        if self.module.params.get('container_log') in BOOLEANS_TRUE:
            # Set the logging path to the /var/log/lxc if uid is root. else
            # set it to the home folder of the user executing.
            try:
                if os.getuid() != 0:
                    log_path = os.getenv('HOME')
                else:
                    if not os.path.isdir('/var/log/lxc/'):
                        os.makedirs('/var/log/lxc/')
                    log_path = '/var/log/lxc/'
            except OSError:
                log_path = os.getenv('HOME')

            build_command.extend([
                '--logfile %s' % os.path.join(
                    log_path, 'lxc-%s.log' % self.container_name
                ),
                '--logpriority %s' % self.module.params.get(
                    'container_log_level'
                ).upper()
            ])

        # Add the template commands to the end of the command if there are any
        template_options = self.module.params.get('template_options', None)
        if template_options:
            build_command.append('-- %s' % template_options)

        rc, return_data, err = self._run_command(build_command)
        if rc != 0:
            message = "Failed executing lxc-create."
            self.failure(
                err=err, rc=rc, msg=message, command=' '.join(build_command)
            )
        else:
            self.state_change = True

    def _container_data(self):
        """Returns a dict of container information.

        :returns: container data
        :rtype: ``dict``
        """

        return {
            'interfaces': self.container.get_interfaces(),
            'ips': self.container.get_ips(),
            'state': self._get_state(),
            'init_pid': int(self.container.init_pid),
            'name': self.container_name,
        }

    def _unfreeze(self):
        """Unfreeze a container.

        :returns: True or False based on if the container was unfrozen.
        :rtype: ``bol``
        """

        unfreeze = self.container.unfreeze()
        if unfreeze:
            self.state_change = True
        return unfreeze

    def _get_state(self):
        """Return the state of a container.

        If the container is not found the state returned is "absent"

        :returns: state of a container as a lower case string.
        :rtype: ``str``
        """

        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            return str(self.container.state).lower()
        return str('absent')

    def _execute_command(self):
        """Execute a shell command."""

        container_command = self.module.params.get('container_command')
        if container_command:
            container_state = self._get_state()
            if container_state == 'frozen':
                self._unfreeze()
            elif container_state == 'stopped':
                self._container_startup()

            self.container.attach_wait(create_script, container_command)
            self.state_change = True

    def _container_startup(self, timeout=60):
        """Ensure a container is started.

        :param timeout: Time before the destroy operation is abandoned.
        :type timeout: ``int``
        """

        self.container = self.get_container_bind()
        for _ in xrange(timeout):
            if self._get_state() != 'running':
                self.container.start()
                self.state_change = True
                # post startup sleep for 1 second.
                time.sleep(1)
            else:
                return True
        self.failure(
            lxc_container=self._container_data(),
            error='Failed to start container'
                  ' [ %s ]' % self.container_name,
            rc=1,
            msg='The container [ %s ] failed to start. Check to lxc is'
                ' available and that the container is in a functional'
                ' state.' % self.container_name
        )

    def _check_archive(self):
        """Create a compressed archive of a container.

        This will store archive_info in as self.archive_info
        """

        if self.module.params.get('archive') in BOOLEANS_TRUE:
            self.archive_info = {
                'archive': self._container_create_tar()
            }

    def _check_clone(self):
        """Create a compressed archive of a container.

        This will store archive_info in as self.archive_info
        """

        clone_name = self.module.params.get('clone_name')
        if clone_name:
            if not self._container_exists(container_name=clone_name, lxc_path=self.lxc_path):
                self.clone_info = {
                    'cloned': self._container_create_clone()
                }
            else:
                self.clone_info = {
                    'cloned': False
                }

    def _destroyed(self, timeout=60):
        """Ensure a container is destroyed.

        :param timeout: Time before the destroy operation is abandoned.
        :type timeout: ``int``
        """

        for _ in xrange(timeout):
            if not self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
                break

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()

            if self._get_state() != 'stopped':
                self.state_change = True
                self.container.stop()

            if self.container.destroy():
                self.state_change = True

            # post destroy attempt sleep for 1 second.
            time.sleep(1)
        else:
            self.failure(
                lxc_container=self._container_data(),
                error='Failed to destroy container'
                      ' [ %s ]' % self.container_name,
                rc=1,
                msg='The container [ %s ] failed to be destroyed. Check'
                    ' that lxc is available and that the container is in a'
                    ' functional state.' % self.container_name
            )

    def _frozen(self, count=0):
        """Ensure a container is frozen.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method='frozen')
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            container_state = self._get_state()
            if container_state == 'frozen':
                pass
            elif container_state == 'running':
                self.container.freeze()
                self.state_change = True
            else:
                self._container_startup()
                self.container.freeze()
                self.state_change = True

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._frozen(count)

    def _restarted(self, count=0):
        """Ensure a container is restarted.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method='restart')
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            if self._get_state() != 'stopped':
                self.container.stop()
                self.state_change = True

            # Run container startup
            self._container_startup()

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._restarted(count)

    def _stopped(self, count=0):
        """Ensure a container is stopped.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method='stop')
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            if self._get_state() != 'stopped':
                self.container.stop()
                self.state_change = True

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._stopped(count)

    def _started(self, count=0):
        """Ensure a container is started.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method='start')
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            container_state = self._get_state()
            if container_state == 'running':
                pass
            elif container_state == 'frozen':
                self._unfreeze()
            elif not self._container_startup():
                self.failure(
                    lxc_container=self._container_data(),
                    error='Failed to start container'
                          ' [ %s ]' % self.container_name,
                    rc=1,
                    msg='The container [ %s ] failed to start. Check to lxc is'
                        ' available and that the container is in a functional'
                        ' state.' % self.container_name
                )

            # Return data
            self._execute_command()

            # Perform any configuration updates
            self._config()

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._started(count)

    def _get_lxc_vg(self):
        """Return the name of the Volume Group used in LXC."""

        build_command = [
            self.module.get_bin_path('lxc-config', True),
            "lxc.bdev.lvm.vg"
        ]
        rc, vg, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='Failed to read LVM VG from LXC config',
                command=' '.join(build_command)
            )
        else:
            return str(vg.strip())

    def _lvm_lv_list(self):
        """Return a list of all lv in a current vg."""

        vg = self._get_lxc_vg()
        build_command = [
            self.module.get_bin_path('lvs', True)
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='Failed to get list of LVs',
                command=' '.join(build_command)
            )

        all_lvms = [i.split() for i in stdout.splitlines()][1:]
        return [lv_entry[0] for lv_entry in all_lvms if lv_entry[1] == vg]

    def _get_vg_free_pe(self, vg_name):
        """Return the available size of a given VG.

        :param vg_name: Name of volume.
        :type vg_name: ``str``
        :returns: size and measurement of an LV
        :type: ``tuple``
        """

        build_command = [
            'vgdisplay',
            vg_name,
            '--units',
            'g'
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to read vg %s' % vg_name,
                command=' '.join(build_command)
            )

        vg_info = [i.strip() for i in stdout.splitlines()][1:]
        free_pe = [i for i in vg_info if i.startswith('Free')]
        _free_pe = free_pe[0].split()
        return float(_free_pe[-2]), _free_pe[-1]

    def _get_lv_size(self, lv_name):
        """Return the available size of a given LV.

        :param lv_name: Name of volume.
        :type lv_name: ``str``
        :returns: size and measurement of an LV
        :type: ``tuple``
        """

        vg = self._get_lxc_vg()
        lv = os.path.join(vg, lv_name)
        build_command = [
            'lvdisplay',
            lv,
            '--units',
            'g'
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to read lv %s' % lv,
                command=' '.join(build_command)
            )

        lv_info = [i.strip() for i in stdout.splitlines()][1:]
        _free_pe = [i for i in lv_info if i.startswith('LV Size')]
        free_pe = _free_pe[0].split()
        return self._roundup(float(free_pe[-2])), free_pe[-1]

    def _lvm_snapshot_create(self, source_lv, snapshot_name,
                             snapshot_size_gb=5):
        """Create an LVM snapshot.

        :param source_lv: Name of lv to snapshot
        :type source_lv: ``str``
        :param snapshot_name: Name of lv snapshot
        :type snapshot_name: ``str``
        :param snapshot_size_gb: Size of snapshot to create
        :type snapshot_size_gb: ``int``
        """

        vg = self._get_lxc_vg()
        free_space, messurement = self._get_vg_free_pe(vg_name=vg)

        if free_space < float(snapshot_size_gb):
            message = (
                'Snapshot size [ %s ] is > greater than [ %s ] on volume group'
                ' [ %s ]' % (snapshot_size_gb, free_space, vg)
            )
            self.failure(
                error='Not enough space to create snapshot',
                rc=2,
                msg=message
            )

        # Create LVM Snapshot
        build_command = [
            self.module.get_bin_path('lvcreate', True),
            "-n",
            snapshot_name,
            "-s",
            os.path.join(vg, source_lv),
            "-L%sg" % snapshot_size_gb
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='Failed to Create LVM snapshot %s/%s --> %s'
                    % (vg, source_lv, snapshot_name)
            )

    def _lvm_lv_mount(self, lv_name, mount_point):
        """mount an lv.

        :param lv_name: name of the logical volume to mount
        :type lv_name: ``str``
        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        vg = self._get_lxc_vg()

        build_command = [
            self.module.get_bin_path('mount', True),
            "/dev/%s/%s" % (vg, lv_name),
            mount_point,
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to mountlvm lv %s/%s to %s'
                    % (vg, lv_name, mount_point)
            )

    def _create_tar(self, source_dir):
        """Create an archive of a given ``source_dir`` to ``output_path``.

        :param source_dir:  Path to the directory to be archived.
        :type source_dir: ``str``
        """

        old_umask = os.umask(int('0077', 8))

        archive_path = self.module.params.get('archive_path')
        if not os.path.isdir(archive_path):
            os.makedirs(archive_path)

        archive_compression = self.module.params.get('archive_compression')
        compression_type = LXC_COMPRESSION_MAP[archive_compression]

        # remove trailing / if present.
        archive_name = '%s.%s' % (
            os.path.join(
                archive_path,
                self.container_name
            ),
            compression_type['extension']
        )

        build_command = [
            self.module.get_bin_path('tar', True),
            '--directory=%s' % os.path.realpath(
                os.path.expanduser(source_dir)
            ),
            compression_type['argument'],
            archive_name,
            '.'
        ]

        rc, stdout, err = self._run_command(
            build_command=build_command,
            unsafe_shell=True
        )

        os.umask(old_umask)

        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to create tar archive',
                command=' '.join(build_command)
            )

        return archive_name

    def _lvm_lv_remove(self, lv_name):
        """Remove an LV.

        :param lv_name: The name of the logical volume
        :type lv_name: ``str``
        """

        vg = self._get_lxc_vg()
        build_command = [
            self.module.get_bin_path('lvremove', True),
            "-f",
            "%s/%s" % (vg, lv_name),
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='Failed to remove LVM LV %s/%s' % (vg, lv_name),
                command=' '.join(build_command)
            )

    def _rsync_data(self, container_path, temp_dir):
        """Sync the container directory to the temp directory.

        :param container_path: path to the container container
        :type container_path: ``str``
        :param temp_dir: path to the temporary local working directory
        :type temp_dir: ``str``
        """
        # This loop is created to support overlayfs archives. This should
        # squash all of the layers into a single archive.
        fs_paths = container_path.split(':')
        if 'overlayfs' in fs_paths:
            fs_paths.pop(fs_paths.index('overlayfs'))

        for fs_path in fs_paths:
            # Set the path to the container data
            fs_path = os.path.dirname(fs_path)

            # Run the sync command
            build_command = [
                self.module.get_bin_path('rsync', True),
                '-aHAX',
                fs_path,
                temp_dir
            ]
            rc, stdout, err = self._run_command(
                build_command,
                unsafe_shell=True
            )
            if rc != 0:
                self.failure(
                    err=err,
                    rc=rc,
                    msg='failed to perform archive',
                    command=' '.join(build_command)
                )

    def _unmount(self, mount_point):
        """Unmount a file system.

        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        build_command = [
            self.module.get_bin_path('umount', True),
            mount_point,
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to unmount [ %s ]' % mount_point,
                command=' '.join(build_command)
            )

    def _overlayfs_mount(self, lowerdir, upperdir, mount_point):
        """mount an lv.

        :param lowerdir: name/path of the lower directory
        :type lowerdir: ``str``
        :param upperdir: name/path of the upper directory
        :type upperdir: ``str``
        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        build_command = [
            self.module.get_bin_path('mount', True),
            '-t overlayfs',
            '-o lowerdir=%s,upperdir=%s' % (lowerdir, upperdir),
            'overlayfs',
            mount_point,
        ]
        rc, stdout, err = self._run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg='failed to mount overlayfs:%s:%s to %s -- Command: %s'
                    % (lowerdir, upperdir, mount_point, build_command)
            )

    def _container_create_tar(self):
        """Create a tar archive from an LXC container.

        The process is as follows:
            * Stop or Freeze the container
            * Create temporary dir
            * Copy container and config to temporary directory
            * If LVM backed:
                * Create LVM snapshot of LV backing the container
                * Mount the snapshot to tmpdir/rootfs
            * Restore the state of the container
            * Create tar of tmpdir
            * Clean up
        """

        # Create a temp dir
        temp_dir = tempfile.mkdtemp()

        # Set the name of the working dir, temp + container_name
        work_dir = os.path.join(temp_dir, self.container_name)

        # LXC container rootfs
        lxc_rootfs = self.container.get_config_item('lxc.rootfs')

        # Test if the containers rootfs is a block device
        block_backed = lxc_rootfs.startswith(os.path.join(os.sep, 'dev'))

        # Test if the container is using overlayfs
        overlayfs_backed = lxc_rootfs.startswith('overlayfs')

        mount_point = os.path.join(work_dir, 'rootfs')

        # Set the snapshot name if needed
        snapshot_name = '%s_lxc_snapshot' % self.container_name

        container_state = self._get_state()
        try:
            # Ensure the original container is stopped or frozen
            if container_state not in ['stopped', 'frozen']:
                if container_state == 'running':
                    self.container.freeze()
                else:
                    self.container.stop()

            # Sync the container data from the container_path to work_dir
            self._rsync_data(lxc_rootfs, temp_dir)

            if block_backed:
                if snapshot_name not in self._lvm_lv_list():
                    if not os.path.exists(mount_point):
                        os.makedirs(mount_point)

                    # Take snapshot
                    size, measurement = self._get_lv_size(
                        lv_name=self.container_name
                    )
                    self._lvm_snapshot_create(
                        source_lv=self.container_name,
                        snapshot_name=snapshot_name,
                        snapshot_size_gb=size
                    )

                    # Mount snapshot
                    self._lvm_lv_mount(
                        lv_name=snapshot_name,
                        mount_point=mount_point
                    )
                else:
                    self.failure(
                        err='snapshot [ %s ] already exists' % snapshot_name,
                        rc=1,
                        msg='The snapshot [ %s ] already exists. Please clean'
                            ' up old snapshot of containers before continuing.'
                            % snapshot_name
                    )
            elif overlayfs_backed:
                lowerdir, upperdir = lxc_rootfs.split(':')[1:]
                self._overlayfs_mount(
                    lowerdir=lowerdir,
                    upperdir=upperdir,
                    mount_point=mount_point
                )

            # Set the state as changed and set a new fact
            self.state_change = True
            return self._create_tar(source_dir=work_dir)
        finally:
            if block_backed or overlayfs_backed:
                # unmount snapshot
                self._unmount(mount_point)

            if block_backed:
                # Remove snapshot
                self._lvm_lv_remove(snapshot_name)

            # Restore original state of container
            if container_state == 'running':
                if self._get_state() == 'frozen':
                    self.container.unfreeze()
                else:
                    self.container.start()

            # Remove tmpdir
            shutil.rmtree(temp_dir)

    def check_count(self, count, method):
        if count > 1:
            self.failure(
                error='Failed to %s container' % method,
                rc=1,
                msg='The container [ %s ] failed to %s. Check to lxc is'
                    ' available and that the container is in a functional'
                    ' state.' % (self.container_name, method)
            )

    def failure(self, **kwargs):
        """Return a Failure when running an Ansible command.

        :param error: ``str``  Error that occurred.
        :param rc: ``int``     Return code while executing an Ansible command.
        :param msg: ``str``    Message to report.
        """

        self.module.fail_json(**kwargs)

    def run(self):
        """Run the main method."""

        action = getattr(self, LXC_ANSIBLE_STATES[self.state])
        action()

        outcome = self._container_data()
        if self.archive_info:
            outcome.update(self.archive_info)

        if self.clone_info:
            outcome.update(self.clone_info)

        self.module.exit_json(
            changed=self.state_change,
            lxc_container=outcome
        )


def main():
    """Ansible Main module."""

    module = AnsibleModule(
        argument_spec=dict(
            name=dict(
                type='str',
                required=True
            ),
            template=dict(
                type='str',
                default='ubuntu'
            ),
            backing_store=dict(
                type='str',
                choices=LXC_BACKING_STORE.keys(),
                default='dir'
            ),
            template_options=dict(
                type='str'
            ),
            config=dict(
                type='path',
            ),
            vg_name=dict(
                type='str',
                default='lxc'
            ),
            thinpool=dict(
                type='str'
            ),
            fs_type=dict(
                type='str',
                default='ext4'
            ),
            fs_size=dict(
                type='str',
                default='5G'
            ),
            directory=dict(
                type='path'
            ),
            zfs_root=dict(
                type='str'
            ),
            lv_name=dict(
                type='str'
            ),
            lxc_path=dict(
                type='path'
            ),
            state=dict(
                choices=LXC_ANSIBLE_STATES.keys(),
                default='started'
            ),
            container_command=dict(
                type='str'
            ),
            container_config=dict(
                type='list',
                elements='str'
            ),
            container_log=dict(
                type='bool',
                default=False
            ),
            container_log_level=dict(
                choices=[n for i in LXC_LOGGING_LEVELS.values() for n in i],
                default='INFO'
            ),
            clone_name=dict(
                type='str',
                required=False
            ),
            clone_snapshot=dict(
                type='bool',
                default='false'
            ),
            archive=dict(
                type='bool',
                default=False
            ),
            archive_path=dict(
                type='path',
            ),
            archive_compression=dict(
                choices=LXC_COMPRESSION_MAP.keys(),
                default='gzip'
            )
        ),
        supports_check_mode=False,
        required_if=([
            ('archive', True, ['archive_path'])
        ]),
    )

    if not HAS_LXC:
        module.fail_json(
            msg='The `lxc` module is not importable. Check the requirements.'
        )

    lv_name = module.params.get('lv_name')
    if not lv_name:
        module.params['lv_name'] = module.params.get('name')

    lxc_manage = LxcContainerManagement(module=module)
    lxc_manage.run()


if __name__ == '__main__':
    main()



#!/usr/bin/python
"""
# Created on Aug 12, 2016
#
# @author: Gaurav Rastogi (grastogi@avinetworks.com) GitHub ID: grastogi23
#
# module_check: not supported
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
"""


from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
---
module: avi_useraccount
author: Chaitanya Deshpande (@chaitanyaavi) <chaitanya.deshpande@avinetworks.com>
short_description: Avi UserAccount Module
description:
    - This module can be used for updating the password of a user.
    - This module is useful for setting up admin password for Controller bootstrap.
requirements: [ avisdk ]
options:
    old_password:
        description:
            - Old password for update password or default password for bootstrap.
    force_change:
        description:
            - If specifically set to true then old password is tried first for controller and then the new password is
              tried. If not specified this flag then the new password is tried first.
        default: false

extends_documentation_fragment:
- community.network.avi

'''

EXAMPLES = '''
  - name: Update user password
    community.network.avi_useraccount:
      controller: ""
      username: ""
      password: new_password
      old_password: ""
      api_version: ""
      force_change: false

  - name: Update user password using avi_credentials
    community.network.avi_useraccount:
      avi_credentials: ""
      old_password: ""
      force_change: false
'''

RETURN = '''
obj:
    description: Avi REST resource
    returned: success, changed
    type: dict
'''

import json
import time
from ansible.module_utils.basic import AnsibleModule
from copy import deepcopy

try:
    from ansible_collections.community.network.plugins.module_utils.network.avi.avi import (
        avi_common_argument_spec, ansible_return, avi_obj_cmp,
        cleanup_absent_fields, HAS_AVI)
    from ansible_collections.community.network.plugins.module_utils.network.avi.avi_api import (
        ApiSession, AviCredentials)
except ImportError:
    HAS_AVI = False


def main():
    argument_specs = dict(
        old_password=dict(type='str', required=True, no_log=True),
        # Flag to specify priority of old/new password while establishing session with controller.
        # To handle both Saas and conventional (Entire state in playbook) scenario.
        force_change=dict(type='bool', default=False)
    )
    argument_specs.update(avi_common_argument_spec())
    module = AnsibleModule(argument_spec=argument_specs)
    if not HAS_AVI:
        return module.fail_json(msg=(
            'Avi python API SDK (avisdk>=17.1) or requests is not installed. '
            'For more details visit https://github.com/avinetworks/sdk.'))
    api_creds = AviCredentials()
    api_creds.update_from_ansible_module(module)
    old_password = module.params.get('old_password')
    force_change = module.params.get('force_change', False)
    data = {
        'old_password': old_password,
        'password': api_creds.password
    }
    # First try old password if 'force_change' is set to true
    if force_change:
        first_pwd = old_password
        second_pwd = api_creds.password
    # First try new password if 'force_change' is set to false or not specified in playbook.
    else:
        first_pwd = api_creds.password
        second_pwd = old_password
    password_changed = False
    try:
        api = ApiSession.get_session(
            api_creds.controller, api_creds.username,
            password=first_pwd, timeout=api_creds.timeout,
            tenant=api_creds.tenant, tenant_uuid=api_creds.tenant_uuid,
            token=api_creds.token, port=api_creds.port)
        if force_change:
            rsp = api.put('useraccount', data=data)
            if rsp:
                password_changed = True
    except Exception:
        pass
    if not password_changed:
        api = ApiSession.get_session(
            api_creds.controller, api_creds.username, password=second_pwd,
            timeout=api_creds.timeout, tenant=api_creds.tenant,
            tenant_uuid=api_creds.tenant_uuid, token=api_creds.token,
            port=api_creds.port)
        if not force_change:
            rsp = api.put('useraccount', data=data)
            if rsp:
                password_changed = True
    if password_changed:
        return ansible_return(module, rsp, True, req=data)
    else:
        return ansible_return(module, rsp, False, req=data)


if __name__ == '__main__':
    main()
